# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.transient_error import TransientError  # noqa: F401,E501
from orcid_api_v3.models.transient_non_empty_string import TransientNonEmptyString  # noqa: F401,E501
from orcid_api_v3.models.url_v30 import UrlV30  # noqa: F401,E501


class ExternalIDV30(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'external_id_type': 'str',
        'external_id_value': 'str',
        'external_id_normalized': 'TransientNonEmptyString',
        'external_id_normalized_error': 'TransientError',
        'external_id_url': 'UrlV30',
        'external_id_relationship': 'str'
    }

    attribute_map = {
        'external_id_type': 'external-id-type',
        'external_id_value': 'external-id-value',
        'external_id_normalized': 'external-id-normalized',
        'external_id_normalized_error': 'external-id-normalized-error',
        'external_id_url': 'external-id-url',
        'external_id_relationship': 'external-id-relationship'
    }

    def __init__(self, external_id_type=None, external_id_value=None, external_id_normalized=None, external_id_normalized_error=None, external_id_url=None, external_id_relationship=None):  # noqa: E501
        """ExternalIDV30 - a model defined in Swagger"""  # noqa: E501
        self._external_id_type = None
        self._external_id_value = None
        self._external_id_normalized = None
        self._external_id_normalized_error = None
        self._external_id_url = None
        self._external_id_relationship = None
        self.discriminator = None
        self.external_id_type = external_id_type
        self.external_id_value = external_id_value
        if external_id_normalized is not None:
            self.external_id_normalized = external_id_normalized
        if external_id_normalized_error is not None:
            self.external_id_normalized_error = external_id_normalized_error
        if external_id_url is not None:
            self.external_id_url = external_id_url
        if external_id_relationship is not None:
            self.external_id_relationship = external_id_relationship

    @property
    def external_id_type(self):
        """Gets the external_id_type of this ExternalIDV30.  # noqa: E501


        :return: The external_id_type of this ExternalIDV30.  # noqa: E501
        :rtype: str
        """
        return self._external_id_type

    @external_id_type.setter
    def external_id_type(self, external_id_type):
        """Sets the external_id_type of this ExternalIDV30.


        :param external_id_type: The external_id_type of this ExternalIDV30.  # noqa: E501
        :type: str
        """
        if external_id_type is None:
            raise ValueError("Invalid value for `external_id_type`, must not be `None`")  # noqa: E501

        self._external_id_type = external_id_type

    @property
    def external_id_value(self):
        """Gets the external_id_value of this ExternalIDV30.  # noqa: E501


        :return: The external_id_value of this ExternalIDV30.  # noqa: E501
        :rtype: str
        """
        return self._external_id_value

    @external_id_value.setter
    def external_id_value(self, external_id_value):
        """Sets the external_id_value of this ExternalIDV30.


        :param external_id_value: The external_id_value of this ExternalIDV30.  # noqa: E501
        :type: str
        """
        if external_id_value is None:
            raise ValueError("Invalid value for `external_id_value`, must not be `None`")  # noqa: E501

        self._external_id_value = external_id_value

    @property
    def external_id_normalized(self):
        """Gets the external_id_normalized of this ExternalIDV30.  # noqa: E501


        :return: The external_id_normalized of this ExternalIDV30.  # noqa: E501
        :rtype: TransientNonEmptyString
        """
        return self._external_id_normalized

    @external_id_normalized.setter
    def external_id_normalized(self, external_id_normalized):
        """Sets the external_id_normalized of this ExternalIDV30.


        :param external_id_normalized: The external_id_normalized of this ExternalIDV30.  # noqa: E501
        :type: TransientNonEmptyString
        """

        self._external_id_normalized = external_id_normalized

    @property
    def external_id_normalized_error(self):
        """Gets the external_id_normalized_error of this ExternalIDV30.  # noqa: E501


        :return: The external_id_normalized_error of this ExternalIDV30.  # noqa: E501
        :rtype: TransientError
        """
        return self._external_id_normalized_error

    @external_id_normalized_error.setter
    def external_id_normalized_error(self, external_id_normalized_error):
        """Sets the external_id_normalized_error of this ExternalIDV30.


        :param external_id_normalized_error: The external_id_normalized_error of this ExternalIDV30.  # noqa: E501
        :type: TransientError
        """

        self._external_id_normalized_error = external_id_normalized_error

    @property
    def external_id_url(self):
        """Gets the external_id_url of this ExternalIDV30.  # noqa: E501


        :return: The external_id_url of this ExternalIDV30.  # noqa: E501
        :rtype: UrlV30
        """
        return self._external_id_url

    @external_id_url.setter
    def external_id_url(self, external_id_url):
        """Sets the external_id_url of this ExternalIDV30.


        :param external_id_url: The external_id_url of this ExternalIDV30.  # noqa: E501
        :type: UrlV30
        """

        self._external_id_url = external_id_url

    @property
    def external_id_relationship(self):
        """Gets the external_id_relationship of this ExternalIDV30.  # noqa: E501


        :return: The external_id_relationship of this ExternalIDV30.  # noqa: E501
        :rtype: str
        """
        return self._external_id_relationship

    @external_id_relationship.setter
    def external_id_relationship(self, external_id_relationship):
        """Sets the external_id_relationship of this ExternalIDV30.


        :param external_id_relationship: The external_id_relationship of this ExternalIDV30.  # noqa: E501
        :type: str
        """
        allowed_values = ["PART_OF", "SELF", "VERSION_OF"]  # noqa: E501
        if external_id_relationship not in allowed_values:
            raise ValueError(
                "Invalid value for `external_id_relationship` ({0}), must be one of {1}"  # noqa: E501
                .format(external_id_relationship, allowed_values)
            )

        self._external_id_relationship = external_id_relationship

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ExternalIDV30, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ExternalIDV30):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.title_v30 import TitleV30  # noqa: F401,E501
from orcid_api_v3.models.translated_title_v30 import TranslatedTitleV30  # noqa: F401,E501


class ResearchResourceTitleV30(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'title': 'TitleV30',
        'translated_title': 'TranslatedTitleV30'
    }

    attribute_map = {
        'title': 'title',
        'translated_title': 'translated-title'
    }

    def __init__(self, title=None, translated_title=None):  # noqa: E501
        """ResearchResourceTitleV30 - a model defined in Swagger"""  # noqa: E501
        self._title = None
        self._translated_title = None
        self.discriminator = None
        if title is not None:
            self.title = title
        if translated_title is not None:
            self.translated_title = translated_title

    @property
    def title(self):
        """Gets the title of this ResearchResourceTitleV30.  # noqa: E501


        :return: The title of this ResearchResourceTitleV30.  # noqa: E501
        :rtype: TitleV30
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this ResearchResourceTitleV30.


        :param title: The title of this ResearchResourceTitleV30.  # noqa: E501
        :type: TitleV30
        """

        self._title = title

    @property
    def translated_title(self):
        """Gets the translated_title of this ResearchResourceTitleV30.  # noqa: E501


        :return: The translated_title of this ResearchResourceTitleV30.  # noqa: E501
        :rtype: TranslatedTitleV30
        """
        return self._translated_title

    @translated_title.setter
    def translated_title(self, translated_title):
        """Sets the translated_title of this ResearchResourceTitleV30.


        :param translated_title: The translated_title of this ResearchResourceTitleV30.  # noqa: E501
        :type: TranslatedTitleV30
        """

        self._translated_title = translated_title

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ResearchResourceTitleV30, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResearchResourceTitleV30):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.activities_summary_v30_rc1 import ActivitiesSummaryV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.history_v30_rc1 import HistoryV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.orcid_identifier_v30_rc1 import OrcidIdentifierV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.person_v30_rc1 import PersonV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.preferences_v30_rc1 import PreferencesV30Rc1  # noqa: F401,E501


class RecordV30Rc1(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'orcid_identifier': 'OrcidIdentifierV30Rc1',
        'preferences': 'PreferencesV30Rc1',
        'history': 'HistoryV30Rc1',
        'person': 'PersonV30Rc1',
        'activities_summary': 'ActivitiesSummaryV30Rc1',
        'path': 'str'
    }

    attribute_map = {
        'orcid_identifier': 'orcid-identifier',
        'preferences': 'preferences',
        'history': 'history',
        'person': 'person',
        'activities_summary': 'activities-summary',
        'path': 'path'
    }

    def __init__(self, orcid_identifier=None, preferences=None, history=None, person=None, activities_summary=None, path=None):  # noqa: E501
        """RecordV30Rc1 - a model defined in Swagger"""  # noqa: E501
        self._orcid_identifier = None
        self._preferences = None
        self._history = None
        self._person = None
        self._activities_summary = None
        self._path = None
        self.discriminator = None
        if orcid_identifier is not None:
            self.orcid_identifier = orcid_identifier
        if preferences is not None:
            self.preferences = preferences
        if history is not None:
            self.history = history
        if person is not None:
            self.person = person
        if activities_summary is not None:
            self.activities_summary = activities_summary
        if path is not None:
            self.path = path

    @property
    def orcid_identifier(self):
        """Gets the orcid_identifier of this RecordV30Rc1.  # noqa: E501


        :return: The orcid_identifier of this RecordV30Rc1.  # noqa: E501
        :rtype: OrcidIdentifierV30Rc1
        """
        return self._orcid_identifier

    @orcid_identifier.setter
    def orcid_identifier(self, orcid_identifier):
        """Sets the orcid_identifier of this RecordV30Rc1.


        :param orcid_identifier: The orcid_identifier of this RecordV30Rc1.  # noqa: E501
        :type: OrcidIdentifierV30Rc1
        """

        self._orcid_identifier = orcid_identifier

    @property
    def preferences(self):
        """Gets the preferences of this RecordV30Rc1.  # noqa: E501


        :return: The preferences of this RecordV30Rc1.  # noqa: E501
        :rtype: PreferencesV30Rc1
        """
        return self._preferences

    @preferences.setter
    def preferences(self, preferences):
        """Sets the preferences of this RecordV30Rc1.


        :param preferences: The preferences of this RecordV30Rc1.  # noqa: E501
        :type: PreferencesV30Rc1
        """

        self._preferences = preferences

    @property
    def history(self):
        """Gets the history of this RecordV30Rc1.  # noqa: E501


        :return: The history of this RecordV30Rc1.  # noqa: E501
        :rtype: HistoryV30Rc1
        """
        return self._history

    @history.setter
    def history(self, history):
        """Sets the history of this RecordV30Rc1.


        :param history: The history of this RecordV30Rc1.  # noqa: E501
        :type: HistoryV30Rc1
        """

        self._history = history

    @property
    def person(self):
        """Gets the person of this RecordV30Rc1.  # noqa: E501


        :return: The person of this RecordV30Rc1.  # noqa: E501
        :rtype: PersonV30Rc1
        """
        return self._person

    @person.setter
    def person(self, person):
        """Sets the person of this RecordV30Rc1.


        :param person: The person of this RecordV30Rc1.  # noqa: E501
        :type: PersonV30Rc1
        """

        self._person = person

    @property
    def activities_summary(self):
        """Gets the activities_summary of this RecordV30Rc1.  # noqa: E501


        :return: The activities_summary of this RecordV30Rc1.  # noqa: E501
        :rtype: ActivitiesSummaryV30Rc1
        """
        return self._activities_summary

    @activities_summary.setter
    def activities_summary(self, activities_summary):
        """Sets the activities_summary of this RecordV30Rc1.


        :param activities_summary: The activities_summary of this RecordV30Rc1.  # noqa: E501
        :type: ActivitiesSummaryV30Rc1
        """

        self._activities_summary = activities_summary

    @property
    def path(self):
        """Gets the path of this RecordV30Rc1.  # noqa: E501


        :return: The path of this RecordV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this RecordV30Rc1.


        :param path: The path of this RecordV30Rc1.  # noqa: E501
        :type: str
        """

        self._path = path

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(RecordV30Rc1, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, RecordV30Rc1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six


class TranslatedTitleV30Rc2(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'value': 'str',
        'language_code': 'str'
    }

    attribute_map = {
        'value': 'value',
        'language_code': 'language-code'
    }

    def __init__(self, value=None, language_code=None):  # noqa: E501
        """TranslatedTitleV30Rc2 - a model defined in Swagger"""  # noqa: E501
        self._value = None
        self._language_code = None
        self.discriminator = None
        if value is not None:
            self.value = value
        self.language_code = language_code

    @property
    def value(self):
        """Gets the value of this TranslatedTitleV30Rc2.  # noqa: E501


        :return: The value of this TranslatedTitleV30Rc2.  # noqa: E501
        :rtype: str
        """
        return self._value

    @value.setter
    def value(self, value):
        """Sets the value of this TranslatedTitleV30Rc2.


        :param value: The value of this TranslatedTitleV30Rc2.  # noqa: E501
        :type: str
        """

        self._value = value

    @property
    def language_code(self):
        """Gets the language_code of this TranslatedTitleV30Rc2.  # noqa: E501


        :return: The language_code of this TranslatedTitleV30Rc2.  # noqa: E501
        :rtype: str
        """
        return self._language_code

    @language_code.setter
    def language_code(self, language_code):
        """Sets the language_code of this TranslatedTitleV30Rc2.


        :param language_code: The language_code of this TranslatedTitleV30Rc2.  # noqa: E501
        :type: str
        """
        if language_code is None:
            raise ValueError("Invalid value for `language_code`, must not be `None`")  # noqa: E501
        allowed_values = ["ab", "aa", "af", "ak", "sq", "am", "ar", "an", "hy", "as", "av", "ae", "ay", "az", "bm", "ba", "eu", "be", "bn", "bh", "bi", "bs", "br", "bg", "my", "ca", "ch", "ce", "zh_CN", "zh_TW", "cu", "cv", "kw", "co", "cr", "hr", "cs", "da", "dv", "nl", "dz", "en", "eo", "et", "ee", "fo", "fj", "fi", "fr", "fy", "ff", "gl", "lg", "ka", "de", "el", "kl", "gn", "gu", "ht", "ha", "iw", "hz", "hi", "ho", "hu", "is", "io", "ig", "in", "ia", "ie", "iu", "ik", "ga", "it", "ja", "jv", "kn", "kr", "ks", "kk", "km", "ki", "rw", "ky", "kv", "kg", "ko", "ku", "kj", "lo", "la", "lv", "li", "ln", "lt", "lu", "lb", "mk", "mg", "ms", "ml", "mt", "gv", "mi", "mr", "mh", "mo", "mn", "na", "nv", "ng", "ne", "nd", "se", "no", "nb", "nn", "ny", "oc", "oj", "or", "om", "os", "pi", "pa", "fa", "pl", "pt", "ps", "qu", "rm", "ro", "rn", "ru", "sm", "sg", "sa", "sc", "gd", "sr", "sn", "ii", "sd", "si", "sk", "sl", "so", "nr", "st", "es", "su", "sw", "ss", "sv", "tl", "ty", "tg", "ta", "tt", "te", "th", "bo", "ti", "to", "ts", "tn", "tr", "tk", "tw", "ug", "uk", "ur", "uz", "ve", "vi", "vo", "wa", "cy", "wo", "xh", "ji", "yo", "za", "zu"]  # noqa: E501
        if language_code not in allowed_values:
            raise ValueError(
                "Invalid value for `language_code` ({0}), must be one of {1}"  # noqa: E501
                .format(language_code, allowed_values)
            )

        self._language_code = language_code

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(TranslatedTitleV30Rc2, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TranslatedTitleV30Rc2):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.completion_date_v30_rc1 import CompletionDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.deactivation_date_v30_rc1 import DeactivationDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.last_modified_date_v30_rc1 import LastModifiedDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.source_v30_rc1 import SourceV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.submission_date_v30_rc1 import SubmissionDateV30Rc1  # noqa: F401,E501


class HistoryV30Rc1(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'creation_method': 'str',
        'completion_date': 'CompletionDateV30Rc1',
        'submission_date': 'SubmissionDateV30Rc1',
        'last_modified_date': 'LastModifiedDateV30Rc1',
        'claimed': 'bool',
        'source': 'SourceV30Rc1',
        'deactivation_date': 'DeactivationDateV30Rc1',
        'verified_email': 'bool',
        'verified_primary_email': 'bool'
    }

    attribute_map = {
        'creation_method': 'creation-method',
        'completion_date': 'completion-date',
        'submission_date': 'submission-date',
        'last_modified_date': 'last-modified-date',
        'claimed': 'claimed',
        'source': 'source',
        'deactivation_date': 'deactivation-date',
        'verified_email': 'verified-email',
        'verified_primary_email': 'verified-primary-email'
    }

    def __init__(self, creation_method=None, completion_date=None, submission_date=None, last_modified_date=None, claimed=None, source=None, deactivation_date=None, verified_email=None, verified_primary_email=None):  # noqa: E501
        """HistoryV30Rc1 - a model defined in Swagger"""  # noqa: E501
        self._creation_method = None
        self._completion_date = None
        self._submission_date = None
        self._last_modified_date = None
        self._claimed = None
        self._source = None
        self._deactivation_date = None
        self._verified_email = None
        self._verified_primary_email = None
        self.discriminator = None
        if creation_method is not None:
            self.creation_method = creation_method
        if completion_date is not None:
            self.completion_date = completion_date
        if submission_date is not None:
            self.submission_date = submission_date
        if last_modified_date is not None:
            self.last_modified_date = last_modified_date
        if claimed is not None:
            self.claimed = claimed
        if source is not None:
            self.source = source
        if deactivation_date is not None:
            self.deactivation_date = deactivation_date
        if verified_email is not None:
            self.verified_email = verified_email
        if verified_primary_email is not None:
            self.verified_primary_email = verified_primary_email

    @property
    def creation_method(self):
        """Gets the creation_method of this HistoryV30Rc1.  # noqa: E501


        :return: The creation_method of this HistoryV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._creation_method

    @creation_method.setter
    def creation_method(self, creation_method):
        """Sets the creation_method of this HistoryV30Rc1.


        :param creation_method: The creation_method of this HistoryV30Rc1.  # noqa: E501
        :type: str
        """
        allowed_values = ["API", "DIRECT", "MEMBER_REFERRED", "WEBSITE", "INTEGRATION_TEST"]  # noqa: E501
        if creation_method not in allowed_values:
            raise ValueError(
                "Invalid value for `creation_method` ({0}), must be one of {1}"  # noqa: E501
                .format(creation_method, allowed_values)
            )

        self._creation_method = creation_method

    @property
    def completion_date(self):
        """Gets the completion_date of this HistoryV30Rc1.  # noqa: E501


        :return: The completion_date of this HistoryV30Rc1.  # noqa: E501
        :rtype: CompletionDateV30Rc1
        """
        return self._completion_date

    @completion_date.setter
    def completion_date(self, completion_date):
        """Sets the completion_date of this HistoryV30Rc1.


        :param completion_date: The completion_date of this HistoryV30Rc1.  # noqa: E501
        :type: CompletionDateV30Rc1
        """

        self._completion_date = completion_date

    @property
    def submission_date(self):
        """Gets the submission_date of this HistoryV30Rc1.  # noqa: E501


        :return: The submission_date of this HistoryV30Rc1.  # noqa: E501
        :rtype: SubmissionDateV30Rc1
        """
        return self._submission_date

    @submission_date.setter
    def submission_date(self, submission_date):
        """Sets the submission_date of this HistoryV30Rc1.


        :param submission_date: The submission_date of this HistoryV30Rc1.  # noqa: E501
        :type: SubmissionDateV30Rc1
        """

        self._submission_date = submission_date

    @property
    def last_modified_date(self):
        """Gets the last_modified_date of this HistoryV30Rc1.  # noqa: E501


        :return: The last_modified_date of this HistoryV30Rc1.  # noqa: E501
        :rtype: LastModifiedDateV30Rc1
        """
        return self._last_modified_date

    @last_modified_date.setter
    def last_modified_date(self, last_modified_date):
        """Sets the last_modified_date of this HistoryV30Rc1.


        :param last_modified_date: The last_modified_date of this HistoryV30Rc1.  # noqa: E501
        :type: LastModifiedDateV30Rc1
        """

        self._last_modified_date = last_modified_date

    @property
    def claimed(self):
        """Gets the claimed of this HistoryV30Rc1.  # noqa: E501


        :return: The claimed of this HistoryV30Rc1.  # noqa: E501
        :rtype: bool
        """
        return self._claimed

    @claimed.setter
    def claimed(self, claimed):
        """Sets the claimed of this HistoryV30Rc1.


        :param claimed: The claimed of this HistoryV30Rc1.  # noqa: E501
        :type: bool
        """

        self._claimed = claimed

    @property
    def source(self):
        """Gets the source of this HistoryV30Rc1.  # noqa: E501


        :return: The source of this HistoryV30Rc1.  # noqa: E501
        :rtype: SourceV30Rc1
        """
        return self._source

    @source.setter
    def source(self, source):
        """Sets the source of this HistoryV30Rc1.


        :param source: The source of this HistoryV30Rc1.  # noqa: E501
        :type: SourceV30Rc1
        """

        self._source = source

    @property
    def deactivation_date(self):
        """Gets the deactivation_date of this HistoryV30Rc1.  # noqa: E501


        :return: The deactivation_date of this HistoryV30Rc1.  # noqa: E501
        :rtype: DeactivationDateV30Rc1
        """
        return self._deactivation_date

    @deactivation_date.setter
    def deactivation_date(self, deactivation_date):
        """Sets the deactivation_date of this HistoryV30Rc1.


        :param deactivation_date: The deactivation_date of this HistoryV30Rc1.  # noqa: E501
        :type: DeactivationDateV30Rc1
        """

        self._deactivation_date = deactivation_date

    @property
    def verified_email(self):
        """Gets the verified_email of this HistoryV30Rc1.  # noqa: E501


        :return: The verified_email of this HistoryV30Rc1.  # noqa: E501
        :rtype: bool
        """
        return self._verified_email

    @verified_email.setter
    def verified_email(self, verified_email):
        """Sets the verified_email of this HistoryV30Rc1.


        :param verified_email: The verified_email of this HistoryV30Rc1.  # noqa: E501
        :type: bool
        """

        self._verified_email = verified_email

    @property
    def verified_primary_email(self):
        """Gets the verified_primary_email of this HistoryV30Rc1.  # noqa: E501


        :return: The verified_primary_email of this HistoryV30Rc1.  # noqa: E501
        :rtype: bool
        """
        return self._verified_primary_email

    @verified_primary_email.setter
    def verified_primary_email(self, verified_primary_email):
        """Sets the verified_primary_email of this HistoryV30Rc1.


        :param verified_primary_email: The verified_primary_email of this HistoryV30Rc1.  # noqa: E501
        :type: bool
        """

        self._verified_primary_email = verified_primary_email

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(HistoryV30Rc1, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, HistoryV30Rc1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.citation import Citation  # noqa: F401,E501
from orcid_api_v3.models.country_v30_rc1 import CountryV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.created_date_v30_rc1 import CreatedDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.external_i_ds_v30_rc1 import ExternalIDsV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.last_modified_date_v30_rc1 import LastModifiedDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.publication_date_v30_rc1 import PublicationDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.source_v30_rc1 import SourceV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.title_v30_rc1 import TitleV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.url_v30_rc1 import UrlV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.work_contributors_v30_rc1 import WorkContributorsV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.work_title_v30_rc1 import WorkTitleV30Rc1  # noqa: F401,E501


class WorkV30Rc1(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'created_date': 'CreatedDateV30Rc1',
        'last_modified_date': 'LastModifiedDateV30Rc1',
        'source': 'SourceV30Rc1',
        'put_code': 'int',
        'path': 'str',
        'title': 'WorkTitleV30Rc1',
        'journal_title': 'TitleV30Rc1',
        'short_description': 'str',
        'citation': 'Citation',
        'type': 'str',
        'publication_date': 'PublicationDateV30Rc1',
        'external_ids': 'ExternalIDsV30Rc1',
        'url': 'UrlV30Rc1',
        'contributors': 'WorkContributorsV30Rc1',
        'language_code': 'str',
        'country': 'CountryV30Rc1',
        'visibility': 'str'
    }

    attribute_map = {
        'created_date': 'created-date',
        'last_modified_date': 'last-modified-date',
        'source': 'source',
        'put_code': 'put-code',
        'path': 'path',
        'title': 'title',
        'journal_title': 'journal-title',
        'short_description': 'short-description',
        'citation': 'citation',
        'type': 'type',
        'publication_date': 'publication-date',
        'external_ids': 'external-ids',
        'url': 'url',
        'contributors': 'contributors',
        'language_code': 'language-code',
        'country': 'country',
        'visibility': 'visibility'
    }

    def __init__(self, created_date=None, last_modified_date=None, source=None, put_code=None, path=None, title=None, journal_title=None, short_description=None, citation=None, type=None, publication_date=None, external_ids=None, url=None, contributors=None, language_code=None, country=None, visibility=None):  # noqa: E501
        """WorkV30Rc1 - a model defined in Swagger"""  # noqa: E501
        self._created_date = None
        self._last_modified_date = None
        self._source = None
        self._put_code = None
        self._path = None
        self._title = None
        self._journal_title = None
        self._short_description = None
        self._citation = None
        self._type = None
        self._publication_date = None
        self._external_ids = None
        self._url = None
        self._contributors = None
        self._language_code = None
        self._country = None
        self._visibility = None
        self.discriminator = None
        if created_date is not None:
            self.created_date = created_date
        if last_modified_date is not None:
            self.last_modified_date = last_modified_date
        if source is not None:
            self.source = source
        if put_code is not None:
            self.put_code = put_code
        if path is not None:
            self.path = path
        if title is not None:
            self.title = title
        if journal_title is not None:
            self.journal_title = journal_title
        if short_description is not None:
            self.short_description = short_description
        if citation is not None:
            self.citation = citation
        if type is not None:
            self.type = type
        if publication_date is not None:
            self.publication_date = publication_date
        if external_ids is not None:
            self.external_ids = external_ids
        if url is not None:
            self.url = url
        if contributors is not None:
            self.contributors = contributors
        if language_code is not None:
            self.language_code = language_code
        if country is not None:
            self.country = country
        if visibility is not None:
            self.visibility = visibility

    @property
    def created_date(self):
        """Gets the created_date of this WorkV30Rc1.  # noqa: E501


        :return: The created_date of this WorkV30Rc1.  # noqa: E501
        :rtype: CreatedDateV30Rc1
        """
        return self._created_date

    @created_date.setter
    def created_date(self, created_date):
        """Sets the created_date of this WorkV30Rc1.


        :param created_date: The created_date of this WorkV30Rc1.  # noqa: E501
        :type: CreatedDateV30Rc1
        """

        self._created_date = created_date

    @property
    def last_modified_date(self):
        """Gets the last_modified_date of this WorkV30Rc1.  # noqa: E501


        :return: The last_modified_date of this WorkV30Rc1.  # noqa: E501
        :rtype: LastModifiedDateV30Rc1
        """
        return self._last_modified_date

    @last_modified_date.setter
    def last_modified_date(self, last_modified_date):
        """Sets the last_modified_date of this WorkV30Rc1.


        :param last_modified_date: The last_modified_date of this WorkV30Rc1.  # noqa: E501
        :type: LastModifiedDateV30Rc1
        """

        self._last_modified_date = last_modified_date

    @property
    def source(self):
        """Gets the source of this WorkV30Rc1.  # noqa: E501


        :return: The source of this WorkV30Rc1.  # noqa: E501
        :rtype: SourceV30Rc1
        """
        return self._source

    @source.setter
    def source(self, source):
        """Sets the source of this WorkV30Rc1.


        :param source: The source of this WorkV30Rc1.  # noqa: E501
        :type: SourceV30Rc1
        """

        self._source = source

    @property
    def put_code(self):
        """Gets the put_code of this WorkV30Rc1.  # noqa: E501


        :return: The put_code of this WorkV30Rc1.  # noqa: E501
        :rtype: int
        """
        return self._put_code

    @put_code.setter
    def put_code(self, put_code):
        """Sets the put_code of this WorkV30Rc1.


        :param put_code: The put_code of this WorkV30Rc1.  # noqa: E501
        :type: int
        """

        self._put_code = put_code

    @property
    def path(self):
        """Gets the path of this WorkV30Rc1.  # noqa: E501


        :return: The path of this WorkV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this WorkV30Rc1.


        :param path: The path of this WorkV30Rc1.  # noqa: E501
        :type: str
        """

        self._path = path

    @property
    def title(self):
        """Gets the title of this WorkV30Rc1.  # noqa: E501


        :return: The title of this WorkV30Rc1.  # noqa: E501
        :rtype: WorkTitleV30Rc1
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this WorkV30Rc1.


        :param title: The title of this WorkV30Rc1.  # noqa: E501
        :type: WorkTitleV30Rc1
        """

        self._title = title

    @property
    def journal_title(self):
        """Gets the journal_title of this WorkV30Rc1.  # noqa: E501


        :return: The journal_title of this WorkV30Rc1.  # noqa: E501
        :rtype: TitleV30Rc1
        """
        return self._journal_title

    @journal_title.setter
    def journal_title(self, journal_title):
        """Sets the journal_title of this WorkV30Rc1.


        :param journal_title: The journal_title of this WorkV30Rc1.  # noqa: E501
        :type: TitleV30Rc1
        """

        self._journal_title = journal_title

    @property
    def short_description(self):
        """Gets the short_description of this WorkV30Rc1.  # noqa: E501


        :return: The short_description of this WorkV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._short_description

    @short_description.setter
    def short_description(self, short_description):
        """Sets the short_description of this WorkV30Rc1.


        :param short_description: The short_description of this WorkV30Rc1.  # noqa: E501
        :type: str
        """

        self._short_description = short_description

    @property
    def citation(self):
        """Gets the citation of this WorkV30Rc1.  # noqa: E501


        :return: The citation of this WorkV30Rc1.  # noqa: E501
        :rtype: Citation
        """
        return self._citation

    @citation.setter
    def citation(self, citation):
        """Sets the citation of this WorkV30Rc1.


        :param citation: The citation of this WorkV30Rc1.  # noqa: E501
        :type: Citation
        """

        self._citation = citation

    @property
    def type(self):
        """Gets the type of this WorkV30Rc1.  # noqa: E501


        :return: The type of this WorkV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this WorkV30Rc1.


        :param type: The type of this WorkV30Rc1.  # noqa: E501
        :type: str
        """
        allowed_values = ["ARTISTIC_PERFORMANCE", "BOOK_CHAPTER", "BOOK_REVIEW", "BOOK", "CONFERENCE_ABSTRACT", "CONFERENCE_PAPER", "CONFERENCE_POSTER", "DATA_SET", "DICTIONARY_ENTRY", "DISCLOSURE", "DISSERTATION", "EDITED_BOOK", "ENCYCLOPEDIA_ENTRY", "INVENTION", "JOURNAL_ARTICLE", "JOURNAL_ISSUE", "LECTURE_SPEECH", "LICENSE", "MAGAZINE_ARTICLE", "MANUAL", "NEWSLETTER_ARTICLE", "NEWSPAPER_ARTICLE", "ONLINE_RESOURCE", "OTHER", "PATENT", "PREPRINT", "REGISTERED_COPYRIGHT", "REPORT", "RESEARCH_TECHNIQUE", "RESEARCH_TOOL", "SOFTWARE", "SPIN_OFF_COMPANY", "STANDARDS_AND_POLICY", "SUPERVISED_STUDENT_PUBLICATION", "TECHNICAL_STANDARD", "TEST", "TRADEMARK", "TRANSLATION", "WEBSITE", "WORKING_PAPER", "UNDEFINED"]  # noqa: E501
        if type not in allowed_values:
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    @property
    def publication_date(self):
        """Gets the publication_date of this WorkV30Rc1.  # noqa: E501


        :return: The publication_date of this WorkV30Rc1.  # noqa: E501
        :rtype: PublicationDateV30Rc1
        """
        return self._publication_date

    @publication_date.setter
    def publication_date(self, publication_date):
        """Sets the publication_date of this WorkV30Rc1.


        :param publication_date: The publication_date of this WorkV30Rc1.  # noqa: E501
        :type: PublicationDateV30Rc1
        """

        self._publication_date = publication_date

    @property
    def external_ids(self):
        """Gets the external_ids of this WorkV30Rc1.  # noqa: E501


        :return: The external_ids of this WorkV30Rc1.  # noqa: E501
        :rtype: ExternalIDsV30Rc1
        """
        return self._external_ids

    @external_ids.setter
    def external_ids(self, external_ids):
        """Sets the external_ids of this WorkV30Rc1.


        :param external_ids: The external_ids of this WorkV30Rc1.  # noqa: E501
        :type: ExternalIDsV30Rc1
        """

        self._external_ids = external_ids

    @property
    def url(self):
        """Gets the url of this WorkV30Rc1.  # noqa: E501


        :return: The url of this WorkV30Rc1.  # noqa: E501
        :rtype: UrlV30Rc1
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this WorkV30Rc1.


        :param url: The url of this WorkV30Rc1.  # noqa: E501
        :type: UrlV30Rc1
        """

        self._url = url

    @property
    def contributors(self):
        """Gets the contributors of this WorkV30Rc1.  # noqa: E501


        :return: The contributors of this WorkV30Rc1.  # noqa: E501
        :rtype: WorkContributorsV30Rc1
        """
        return self._contributors

    @contributors.setter
    def contributors(self, contributors):
        """Sets the contributors of this WorkV30Rc1.


        :param contributors: The contributors of this WorkV30Rc1.  # noqa: E501
        :type: WorkContributorsV30Rc1
        """

        self._contributors = contributors

    @property
    def language_code(self):
        """Gets the language_code of this WorkV30Rc1.  # noqa: E501


        :return: The language_code of this WorkV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._language_code

    @language_code.setter
    def language_code(self, language_code):
        """Sets the language_code of this WorkV30Rc1.


        :param language_code: The language_code of this WorkV30Rc1.  # noqa: E501
        :type: str
        """

        self._language_code = language_code

    @property
    def country(self):
        """Gets the country of this WorkV30Rc1.  # noqa: E501


        :return: The country of this WorkV30Rc1.  # noqa: E501
        :rtype: CountryV30Rc1
        """
        return self._country

    @country.setter
    def country(self, country):
        """Sets the country of this WorkV30Rc1.


        :param country: The country of this WorkV30Rc1.  # noqa: E501
        :type: CountryV30Rc1
        """

        self._country = country

    @property
    def visibility(self):
        """Gets the visibility of this WorkV30Rc1.  # noqa: E501


        :return: The visibility of this WorkV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        """Sets the visibility of this WorkV30Rc1.


        :param visibility: The visibility of this WorkV30Rc1.  # noqa: E501
        :type: str
        """
        allowed_values = ["LIMITED", "REGISTERED_ONLY", "PUBLIC", "PRIVATE"]  # noqa: E501
        if visibility not in allowed_values:
            raise ValueError(
                "Invalid value for `visibility` ({0}), must be one of {1}"  # noqa: E501
                .format(visibility, allowed_values)
            )

        self._visibility = visibility

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(WorkV30Rc1, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, WorkV30Rc1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



"""HUB API."""

from datetime import datetime
import re
from urllib.parse import unquote, urlencode
import dateutil.parser

import jsonschema
import requests
import validators
import yaml
from uuid import UUID
from flask import (Response, abort, current_app, jsonify, make_response, render_template, request,
                   stream_with_context, url_for)
from flask.views import MethodView
from flask_login import current_user, login_user
from flask_restful import Resource, reqparse
from flask_swagger import swagger

from . import api, app, db, models, oauth, schemas
from .login_provider import roles_required
from .models import (ORCID_ID_REGEX, AffiliationRecord, Client, FundingRecord, OrcidToken,
                     PeerReviewRecord, PropertyRecord, Role, Task, TaskType, User, UserOrg,
                     validate_orcid_id, WorkRecord)
from .utils import dump_yaml, is_valid_url, register_orcid_webhook, enqueue_task_records

ORCID_API_VERSION_REGEX = re.compile(r"^v[2-3].\d+(_rc\d+)?$")
SCOPE_REGEX = re.compile(r"^(/[a-z\-]+)+(\,(/[a-z\-]+)+)*$")


def prefers_yaml():
    """Test if the client prefers YAML."""
    best = request.accept_mimetypes.best_match(["text/yaml", "application/x-yaml"])
    return (best in [
        "text/yaml",
        "application/x-yaml",
    ] and request.accept_mimetypes[best] > request.accept_mimetypes["application/json"])


def validate_uuid4(uuid_string):
    """Validate that a UUID string is in fact a valid uuid4.

    Happily, the uuid module does the actual checking for us.
    It is vital that the 'version' kwarg be passed
    to the UUID() call, otherwise any 32-character
    hex string is considered valid.
    """
    try:
        val = UUID(uuid_string, version=4)
    except ValueError:
        return False
    return val.hex == uuid_string.replace('-', '')


@app.route('/api/me')
@app.route("/api/v1.0/me")
@oauth.require_oauth()
def me():
    """Get the token user data."""
    user = request.oauth.user
    return jsonify(email=user.email, name=user.name)


class AppResource(Resource):
    """Common application resource."""

    decorators = [
        oauth.require_oauth(),
    ]

    def dispatch_request(self, *args, **kwargs):
        """Do some pre-handling and post-handling."""
        resp = super().dispatch_request(*args, **kwargs)
        if isinstance(resp, tuple) and len(resp) == 2:
            resp, status_code = resp
            assert isinstance(resp, Response) and isinstance(status_code, int)
            resp.status_code = status_code
            return resp
        return resp

    def httpdate(self, dt):
        """Return a string representation of a date according to RFC 1123 (HTTP/1.1).

        The supplied date must be in UTC.
        """
        weekday = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][dt.weekday()]
        month = [
            "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
        ][dt.month - 1]
        return "%s, %02d %s %04d %02d:%02d:%02d GMT" % (weekday, dt.day, month, dt.year, dt.hour,
                                                        dt.minute, dt.second)

    @property
    def is_yaml_request(self):
        """Test if the request body content type is YAML."""
        return request.content_type in ["text/yaml", "application/x-yaml"]

    def handle_user(self, identifier=None):
        """Create, update or delete user account entry."""
        login_user(request.oauth.user)

        if request.method != "DELETE":
            if self.is_yaml_request:
                try:
                    data = yaml.load(request.data)
                except Exception as ex:
                    return jsonify({
                        "error": "Invalid request format. Only JSON or YAML are acceptable.",
                        "message": str(ex)
                    }), 415
            else:
                data = request.get_json()
            if not data:
                return jsonify({"error": "Invalid request format. Only JSON or YAML are acceptable."}), 415
            try:
                if request.method != "PATCH":
                    jsonschema.validate(data, schemas.hub_user)
            except jsonschema.exceptions.ValidationError as ex:
                return jsonify({"error": "Validation error.", "message": ex.message}), 422
            except Exception as ex:
                return jsonify({"error": "Unhandled exception occurred.", "exception": ex}), 400

            org = current_user.organisation
            email = data.get("email")
            if email and not validators.email(email):
                return jsonify({"error": "Validation error.", "message": f"Invalid email address: {email}"}), 422
            orcid = data.get("orcid")
            if orcid:
                try:
                    models.validate_orcid_id(orcid)
                except Exception as ex:
                    return jsonify({"error": f"Incorrect ORCID iD value '{orcid}': {ex}"}), 422

        created = False
        if identifier:
            if validators.email(identifier):
                user = User.select().where(User.email == identifier).first()
            elif ORCID_ID_REGEX.match(identifier):
                try:
                    models.validate_orcid_id(identifier)
                except Exception as ex:
                    return jsonify({"error": f"Incorrect identifier value '{identifier}': {ex}"}), 400
                user = User.select().where(User.orcid == identifier).first()
            else:
                return jsonify({"error": f"Incorrect identifier value: {identifier}."}), 400
            if not user:
                return jsonify({"error": f"The user ({identifier}) doesn't exist."}), 404

        elif request.method == "POST":
            user = User.select()
            if email:
                user = user.orwhere(User.email == email)
            if orcid:
                user = user.orwhere(User.orcid == orcid)
            user = user.first()
            if not user:
                user, created = User.get_or_create(email=email, orcid=orcid)
            if not created and (user.organisation != org or not UserOrg.select().where(
                    UserOrg.org == org, UserOrg.user == user).exists()):
                return jsonify({"error": f"The user ({identifier or email or orcid}) already exists."}), 400
            if created:
                user.organisation = org
                UserOrg.create(user=user, org=org)

        if request.method == "DELETE":
            try:
                user.delete_instance()
                return make_response('', 204)
            except Exception as ex:
                return jsonify({"error": "Failed to delete the user.", "message": str(ex)}), 400

        else:
            for k in ["orcid", "email", "name", "eppn", "confirmed"]:
                if k in data:
                    setattr(user, k, data[k])
            try:
                user.save()
            except Exception as ex:
                return jsonify({"error": "Failed to update the user.", "message": str(ex)}), 400

            data = user.to_dict(
                    recurse=False, to_dashes=True,
                    only=[User.email, User.eppn, User.name, User.orcid, User.confirmed, User.updated_at])
            resp = yamlfy(data) if prefers_yaml() else jsonify(data)
            resp.headers["Last-Modified"] = self.httpdate(user.updated_at or user.created_at)

            resp.status_code = 201 if created else 200
            return resp


def changed_path(name, value):
    """Create query string with a new parameter value."""
    link = request.path
    if request.args:
        link += '&' + urlencode(
            {k: v if k != name else value
                for k, v in request.args.items()})
    return link


class AppResourceList(AppResource):
    """Resource list resource with helpers for pagination."""

    @models.lazy_property
    def page(self):
        """Get the current queried page."""
        try:
            return int(request.args.get("page", 1))
        except:
            return 1

    @models.lazy_property
    def page_size(self):  # noqa: D402
        """Get the current query page size, default: 20."""
        try:
            return int(request.args.get("page_size", 20))
        except:
            return 20

    @models.lazy_property
    def next_link(self):
        """Get the next page link of the requested resource."""
        return changed_path("page", self.page + 1)

    @models.lazy_property
    def previous_link(self):
        """Get the previous page link of the requested resource."""
        if self.page <= 1:
            return
        return changed_path("page", self.page - 1)

    @models.lazy_property
    def first_link(self):
        """Get the first page link of the requested resource."""
        return changed_path("page", 1)

    def api_response(self, query, exclude=None, only=None):
        """Create and return API response with pagination links."""
        query = query.paginate(self.page, self.page_size)
        records = [r.to_dict(recurse=False, to_dashes=True, exclude=exclude, only=only) for r in query]
        resp = yamlfy(records) if prefers_yaml() else jsonify(records)
        resp.headers["Pagination-Page"] = self.page
        resp.headers["Pagination-Page-Size"] = self.page_size
        resp.headers["Pagination-Count"] = len(records)
        resp.headers["Link"] = f'<{request.full_path}>;rel="self"'
        if self.page != 1:
            resp.headers["Link"] += f', <{self.first_link}>;rel="first"'
        if self.previous_link:
            resp.headers["Link"] += f', <{self.previous_link}>;rel="prev"'
        if len(records) == self.page_size and self.next_link:
            resp.headers["Link"] += f', <{self.next_link}>;rel="next"'
        return resp


class TaskResource(AppResource):
    """Common task related resource."""

    available_task_types = [t.name for t in TaskType]

    def dispatch_request(self, *args, **kwargs):
        """Do some pre-handling..."""
        parser = reqparse.RequestParser()
        parser.add_argument(
            "type", type=str, required=False,
            help="Task type: " + ", ".join(self.available_task_types))
        parser.add_argument("filename", type=str, help="Filename of the task.")
        # TODO: fix Flask-Restful
        try:
            parsed_args = parser.parse_args()
            task_type = parsed_args.get("type")
            if task_type:
                task_type = task_type.upper()
            filename = request.args.get("filename") or parsed_args.get("filename")
        # TODO: fix Flask-Restful
        # TODO: Remove when the fix gets merged in
        except ValueError:
            filename, task_type = request.args.get("filename"), None
        self.filename = filename
        if task_type and any(task_type == t.name for t in TaskType):
            self.task_type = TaskType[task_type]
        else:
            self.task_type = None
        return super().dispatch_request(*args, **kwargs)

    def jsonify_task(self, task):
        """Create JSON response with the task payload."""
        if isinstance(task, int):
            login_user(request.oauth.user)
            try:
                task = Task.get(id=task)
            except Task.DoesNotExist:
                return jsonify({"error": "The task doesn't exist."}), 404
            if task.created_by != current_user:
                return jsonify({"error": "Access denied."}), 403
        if request.method != "HEAD":
            if task.task_type in [
                    TaskType.AFFILIATION, TaskType.FUNDING, TaskType.PEER_REVIEW,
                    TaskType.PROPERTY, TaskType.WORK, TaskType.OTHER_ID
            ]:
                resp = jsonify(task.to_export_dict())
            else:
                raise Exception(f"Suppor for {task} has not yet been implemented.")
        else:
            resp = make_response('')
        resp.headers["Last-Modified"] = self.httpdate(task.updated_at or task.created_at)
        return resp

    def delete_task(self, task_id):
        """Delete the task."""
        login_user(request.oauth.user)
        try:
            task = Task.get(id=task_id)
        except Task.DoesNotExist:
            return jsonify({"error": "The task doesn't exist."}), 404
        except Exception as ex:
            app.logger.exception(f"Failed to find the task with ID: {task_id}")
            return jsonify({"error": "Unhandled exception occurred.", "exception": str(ex)}), 400

        if task.created_by != current_user:
            abort(403)
        task.delete_instance()
        return {"message": "The task was successfully deleted."}

    def handle_affiliation_task(self, task_id=None):
        """Handle PUT, POST, or PATCH request. Request body expected to be encoded in JSON."""
        login_user(request.oauth.user)

        if self.is_yaml_request:
            try:
                data = yaml.load(request.data)
            except Exception as ex:
                return jsonify({
                    "error": "Invalid request format. Only JSON, CSV, or TSV are acceptable.",
                    "message": str(ex)
                }), 415
        else:
            data = request.get_json()

        if not data:
            return jsonify({"error": "Invalid request format. Only JSON, CSV, or TSV are acceptable."}), 415
        try:
            if request.method != "PATCH":
                jsonschema.validate(data, schemas.affiliation_task)
        except jsonschema.exceptions.ValidationError as ex:
            return jsonify({"error": "Validation error.", "message": ex.message}), 422
        except Exception as ex:
            return jsonify({"error": "Unhandled exception occurred.", "exception": ex}), 400
        if "records" not in data:
            return jsonify({"error": "Validation error.", "message": "Missing affiliation records."}), 422

        filename = (data.get("filename") or self.filename or datetime.utcnow().isoformat(timespec="seconds"))
        if task_id:
            try:
                task = Task.get(id=task_id)
            except Task.DoesNotExist:
                return jsonify({"error": "The task doesn't exist."}), 404
            if task.created_by != current_user:
                return jsonify({"error": "Access denied."}), 403
        try:
            task = AffiliationRecord.load(
                data,
                filename=filename,
                task_id=task_id,
                skip_schema_validation=True,
                override=(request.method == "POST"))
        except Exception as ex:
            db.rollback()
            app.logger.exception("Failed to handle affiliation API request.")
            return jsonify({"error": "Unhandled exception occurred.", "exception": str(ex)}), 400

        return self.jsonify_task(task)

    def handle_task(self, task_id=None):
        """Handle PUT, POST, or PATCH request. Request body expected to be encoded in JSON."""
        try:
            login_user(request.oauth.user)
            if task_id:
                try:
                    task = Task.get(id=task_id)
                except Task.DoesNotExist:
                    return jsonify({"error": "The task doesn't exist."}), 404
                if task.created_by != current_user:
                    return jsonify({"error": "Access denied."}), 403
            else:
                task = None
            task = self.load_from_json(task=task)
            enqueue_task_records(task)
        except Exception as ex:
            app.logger.exception("Failed to handle funding API request.")
            return jsonify({"error": "Unhandled exception occurred.", "exception": str(ex)}), 400

        return self.jsonify_task(task)

    def head(self, task_id):
        """Handle HEAD request.

        ---
        summary: "Return task update time-stamp."
        description: "Return record processing task update time-stamp."
        parameters:
          - name: "task_id"
            in: "path"
            description: "Task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)


class TaskList(TaskResource, AppResourceList):
    """Task list services."""

    def get(self, *args, **kwargs):
        """
        Retrieve the list of all submitted task.

        ---
        tags:
          - "tasks"
        summary: "Retrieve the list of all submitted task."
        description: "Retrieve the list of all submitted task."
        produces:
          - "application/json"
          - "text/yaml"
        definitions:
        - schema:
            id: Task
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - AFFILIATION
                - FUNDING
                - PEER_REVIEW
                - PROPERTY
                - WORK
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
        parameters:
          - name: "type"
            in: "query"
            required: false
            description: "The task type: AFFILIATION, FUNDING, etc."
            type: "string"
            enum:
              - AFFILIATION
              - FUNDING
              - PEER_REVIEW
              - PROPERTY
              - WORK
          - in: query
            name: page
            description: The number of the page of retrieved data starting counting from 1
            type: integer
            minimum: 0
            default: 1
          - in: query
            name: page_size
            description: The size of the data page
            type: integer
            minimum: 0
            default: 20
        responses:
          200:
            description: "successful operation"
            schema:
              type: array
              items:
                $ref: "#/definitions/Task"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        login_user(request.oauth.user)
        query = Task.select().where(Task.org_id == current_user.organisation_id)
        task_type = request.args.get("type")
        if task_type:
            query = query.where(Task.task_type == TaskType[task_type.upper()].value)
        return self.api_response(query, exclude=[Task.created_by, Task.updated_by, Task.org])


class AffiliationListAPI(TaskResource):
    """Affiliation list API."""

    def post(self, *args, **kwargs):
        """Upload the affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Post the affiliation list task."
        description: "Post the affiliation list task."
        consumes:
        - application/json
        - text/csv
        - text/yaml
        produces:
        - application/json
        parameters:
        - name: "filename"
          required: false
          in: "query"
          description: "The batch process filename."
          type: "string"
        - name: body
          in: body
          description: "Affiliation task."
          schema:
            $ref: "#/definitions/AffiliationTask"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/AffiliationTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        definitions:
        - schema:
            id: AffiliationTask
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - AFFILIATION
                - FUNDING
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
              records:
                type: array
                items:
                  $ref: "#/definitions/AffiliationTaskRecord"
        - schema:
            id: AffiliationTaskRecord
            required:
              - email
              - first-name
              - last-name
              - affiliation-type
            properties:
              id:
                type: integer
                format: int64
              put-code:
                type: string
              external-id:
                type: string
              is-active:
                type: boolean
              email:
                type: string
              first-name:
                type: string
              last-name:
                type: string
              role:
                type: string
              organisation:
                type: string
              department:
                type: string
              city:
                type: string
              state:
                type: string
              country:
                type: string
              disambiguated-id:
                type: string
              disambiguated-source:
                type: string
              affiliation-type:
                type: string
                enum:
                  - student
                  - staff
              start-date:
                type: string
              end-date:
                type: string
              processed-at:
                type: string
                format: date-time
              status:
                type: string
              orcid:
                type: string
                format: "^[0-9]{4}-?[0-9]{4}-?[0-9]{4}-?[0-9]{4}$"
                description: "User ORCID ID"
        """
        login_user(request.oauth.user)
        if request.content_type in ["text/csv", "text/tsv"]:
            task = Task.load_from_csv(request.data.decode("utf-8"), filename=self.filename)
            return self.jsonify_task(task)
        return self.handle_affiliation_task()


class AffiliationAPI(TaskResource):
    """Affiliation task services."""

    def get(self, task_id):
        """
        Retrieve the specified affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Retrieve the specified affiliation task."
        description: "Retrieve the specified affiliation task."
        produces:
          - "application/json"
        parameters:
          - name: "task_id"
            required: true
            in: "path"
            description: "Affiliation task ID."
            type: "integer"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/AffiliationTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)

    def post(self, task_id):
        """Upload the task and completely override the affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Update the affiliation task."
        description: "Update the affiliation task."
        consumes:
          - application/json
          - text/yaml
        definitions:
        parameters:
          - name: "task_id"
            in: "path"
            description: "Affiliation task ID."
            required: true
            type: "integer"
          - in: body
            name: affiliationTask
            description: "Affiliation task."
            schema:
              $ref: "#/definitions/AffiliationTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/AffiliationTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_affiliation_task(task_id)

    def put(self, task_id):
        """Update the affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Update the affiliation task."
        description: "Update the affiliation task."
        consumes:
          - application/json
          - text/yaml
        parameters:
          - name: "task_id"
            in: "path"
            description: "Affiliation task ID."
            required: true
            type: "integer"
          - in: body
            name: affiliationTask
            description: "Affiliation task."
            schema:
              $ref: "#/definitions/AffiliationTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/AffiliationTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_affiliation_task(task_id)

    def patch(self, task_id):
        """Update the affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Update the affiliation task."
        description: "Update the affiliation task."
        consumes:
          - application/json
          - text/yaml
        parameters:
          - name: "task_id"
            in: "path"
            description: "Affiliation task ID."
            required: true
            type: "integer"
          - in: body
            name: affiliationTask
            description: "Affiliation task."
            schema:
              $ref: "#/definitions/AffiliationTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/AffiliationTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_affiliation_task(task_id)

    def delete(self, task_id):
        """Delete the specified affiliation task.

        ---
        tags:
          - "affiliations"
        summary: "Delete the specified affiliation task."
        description: "Delete the specified affiliation task."
        parameters:
          - name: "task_id"
            in: "path"
            description: "Affiliation task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.delete_task(task_id)


api.add_resource(TaskList, "/api/v1.0/tasks")
api.add_resource(AffiliationListAPI, "/api/v1.0/affiliations")
api.add_resource(AffiliationAPI, "/api/v1.0/affiliations/<int:task_id>")


class FundListAPI(TaskResource):
    """Fund list API."""

    def load_from_json(self, task=None):
        """Load Funding records form the JSON upload."""
        return FundingRecord.load_from_json(
            request.data.decode("utf-8"), filename=self.filename, task=task)

    def post(self, *args, **kwargs):
        """Upload the fund task.

        ---
        tags:
          - "funds"
        summary: "Post the fund list task."
        description: "Post the fund list task."
        consumes:
        - application/json
        - text/csv
        - text/yaml
        produces:
        - application/json
        parameters:
        - name: "filename"
          required: false
          in: "query"
          description: "The batch process filename."
          type: "string"
        - name: body
          in: body
          description: "Fund task."
          schema:
            $ref: "#/definitions/FundTask"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/FundTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        definitions:
        - schema:
            id: FundTask
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - fund
                - FUNDING
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
              records:
                type: array
                items:
                  $ref: "#/definitions/FundTaskRecord"
        - schema:
            id: FundTaskRecord
            type: object
        """
        login_user(request.oauth.user)
        if request.content_type in ["text/csv", "text/tsv"]:
            task = FundingRecord.load_from_csv(request.data.decode("utf-8"), filename=self.filename)
            return self.jsonify_task(task)
        return self.handle_task()


class FundAPI(FundListAPI):
    """Fund task services."""

    def get(self, task_id):
        """
        Retrieve the specified fund task.

        ---
        tags:
          - "funds"
        summary: "Retrieve the specified fund task."
        description: "Retrieve the specified fund task."
        produces:
          - "application/json"
        parameters:
          - name: "task_id"
            required: true
            in: "path"
            description: "Fund task ID."
            type: "integer"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/FundTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)

    def post(self, task_id):
        """Upload the task and completely override the fund task.

        ---
        tags:
          - "funds"
        summary: "Update the fund task."
        description: "Update the fund task."
        consumes:
          - application/json
          - text/yaml
        definitions:
        parameters:
          - name: "task_id"
            in: "path"
            description: "Fund task ID."
            required: true
            type: "integer"
          - in: body
            name: fundTask
            description: "Fund task."
            schema:
              $ref: "#/definitions/FundTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/FundTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def delete(self, task_id):
        """Delete the specified fund task.

        ---
        tags:
          - "funds"
        summary: "Delete the specified fund task."
        description: "Delete the specified fund task."
        parameters:
          - name: "task_id"
            in: "path"
            description: "Fund task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.delete_task(task_id)


api.add_resource(FundListAPI, "/api/v1.0/funds")
api.add_resource(FundAPI, "/api/v1.0/funds/<int:task_id>")


class WorkListAPI(TaskResource):
    """Work list API."""

    def load_from_json(self, task=None):
        """Load Working records form the JSON upload."""
        return WorkRecord.load_from_json(
            request.data.decode("utf-8"), filename=self.filename, task=task)

    def post(self, *args, **kwargs):
        """Upload the work record processing task.

        ---
        tags:
          - "works"
        summary: "Post the work list task."
        description: "Post the work record processing task."
        consumes:
        - application/json
        - text/yaml
        produces:
        - application/json
        parameters:
        - name: "filename"
          required: false
          in: "query"
          description: "The batch process filename."
          type: "string"
        - name: body
          in: body
          description: "Work task."
          schema:
            $ref: "#/definitions/WorkTask"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/WorkTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        definitions:
        - schema:
            id: WorkTask
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - WORK
                - WORKING
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
              records:
                type: array
                items:
                  $ref: "#/definitions/WorkTaskRecord"
        - schema:
            id: WorkTaskRecord
            type: object
        """
        login_user(request.oauth.user)
        if request.content_type in ["text/csv", "text/tsv"]:
            task = WorkRecord.load_from_csv(request.data.decode("utf-8"), filename=self.filename)
            return self.jsonify_task(task)
        return self.handle_task()


class WorkAPI(WorkListAPI):
    """Work record processing task services."""

    def get(self, task_id):
        """
        Retrieve the specified work record processing task.

        ---
        tags:
          - "works"
        summary: "Retrieve the specified work task."
        description: "Retrieve the specified work record processing task."
        produces:
          - "application/json"
        parameters:
          - name: "task_id"
            required: true
            in: "path"
            description: "Work task ID."
            type: "integer"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/WorkTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)

    def post(self, task_id):
        """Upload the task and completely override the work record processing task.

        ---
        tags:
          - "works"
        summary: "Update the work task."
        description: "Update the work record processing task."
        consumes:
          - application/json
          - text/yaml
        definitions:
        parameters:
          - name: "task_id"
            in: "path"
            description: "Work task ID."
            required: true
            type: "integer"
          - in: body
            name: workTask
            description: "Work task."
            schema:
              $ref: "#/definitions/WorkTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/WorkTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def delete(self, task_id):
        """Delete the specified work task.

        ---
        tags:
          - "works"
        summary: "Delete the specified work task."
        description: "Delete the specified work record processing task."
        parameters:
          - name: "task_id"
            in: "path"
            description: "Work task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.delete_task(task_id)


api.add_resource(WorkListAPI, "/api/v1.0/works")
api.add_resource(WorkAPI, "/api/v1.0/works/<int:task_id>")


class PeerReviewListAPI(TaskResource):
    """PeerReview list API."""

    def load_from_json(self, task=None):
        """Load PeerReviewing records form the JSON upload."""
        return PeerReviewRecord.load_from_json(
            request.data.decode("utf-8"), filename=self.filename, task=task)

    def post(self, *args, **kwargs):
        """Upload the peer review record processing task.

        ---
        tags:
          - "peer-reviews"
        summary: "Post the peer review list task."
        description: "Post the peer review record processing task."
        consumes:
        - application/json
        - text/yaml
        produces:
        - application/json
        parameters:
        - name: "filename"
          required: false
          in: "query"
          description: "The batch process filename."
          type: "string"
        - name: body
          in: body
          description: "PeerReview task."
          schema:
            $ref: "#/definitions/PeerReviewTask"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PeerReviewTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        definitions:
        - schema:
            id: PeerReviewTask
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - peer_review
                - peer_reviewING
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
              records:
                type: array
                items:
                  $ref: "#/definitions/PeerReviewTaskRecord"
        - schema:
            id: PeerReviewTaskRecord
            type: object
        """
        login_user(request.oauth.user)
        if request.content_type in ["text/csv", "text/tsv"]:
            task = PeerReviewRecord.load_from_csv(request.data.decode("utf-8"), filename=self.filename)
            return self.jsonify_task(task)
        return self.handle_task()


class PeerReviewAPI(PeerReviewListAPI):
    """PeerReview record processing task services."""

    def get(self, task_id):
        """
        Retrieve the specified peer review record processing task.

        ---
        tags:
          - "peer-reviews"
        summary: "Retrieve the specified peer review task."
        description: "Retrieve the specified peer review record processing task."
        produces:
          - "application/json"
        parameters:
          - name: "task_id"
            required: true
            in: "path"
            description: "PeerReview task ID."
            type: "integer"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PeerReviewTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)

    def post(self, task_id):
        """Upload the task and completely override the peer review record processing task.

        ---
        tags:
          - "peer-reviews"
        summary: "Update the peer review task."
        description: "Update the peer review record processing task."
        consumes:
          - application/json
          - text/yaml
        definitions:
        parameters:
          - name: "task_id"
            in: "path"
            description: "PeerReview task ID."
            required: true
            type: "integer"
          - in: body
            name: peerReviewTask
            description: "PeerReview task."
            schema:
              $ref: "#/definitions/PeerReviewTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PeerReviewTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def delete(self, task_id):
        """Delete the specified peer-review task.

        ---
        tags:
          - "peer-reviews"
        summary: "Delete the specified peer review task."
        description: "Delete the specified peer review record processing task."
        parameters:
          - name: "task_id"
            in: "path"
            description: "PeerReview task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.delete_task(task_id)


api.add_resource(PeerReviewListAPI, "/api/v1.0/peer-reviews")
api.add_resource(PeerReviewAPI, "/api/v1.0/peer-reviews/<int:task_id>")


class PropertyListAPI(TaskResource):
    """Property list API."""

    def load_from_json(self, task=None):
        """Load records form the JSON upload."""
        return PropertyRecord.load_from_json(
            request.data.decode("utf-8"), filename=self.filename, task=task)

    def post(self, *args, **kwargs):
        """Upload the property task.

        ---
        tags:
          - "properties"
        summary: "Post the property list task."
        description: "Post the property list task."
        consumes:
        - application/json
        - text/csv
        - text/yaml
        produces:
        - application/json
        parameters:
        - name: "filename"
          required: false
          in: "query"
          description: "The batch process filename."
          type: "string"
        - name: body
          in: body
          description: "Property task."
          schema:
            $ref: "#/definitions/PropertyTask"
        responses:
          200:
            description: "Successful operation"
            schema:
              $ref: "#/definitions/PropertyTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        definitions:
        - schema:
            id: PropertyTask
            properties:
              id:
                type: integer
                format: int64
              filename:
                type: string
              task-type:
                type: string
                enum:
                - PROPERTY
              created-at:
                type: string
                format: date-time
              expires-at:
                type: string
                format: date-time
              completed-at:
                type: string
                format: date-time
              records:
                type: array
                items:
                  $ref: "#/definitions/PropertyTaskRecord"
        - schema:
            id: PropertyTaskRecord
            required:
              - type
              - value
            properties:
              id:
                type: integer
                format: int64
              put-code:
                type: string
              type:
                type: string
                enum:
                  - COUNTRY
                  - EMAIL
                  - ID
                  - KEYWORD
                  - NAME
                  - URL
                  - WEBSITE
              name:
                type: string
              value:
                type: string
              external-id:
                type: string
              is-active:
                type: boolean
              email:
                type: string
              first-name:
                type: string
              last-name:
                type: string
              role:
                type: string
              organisation:
                type: string
              department:
                type: string
              city:
                type: string
              state:
                type: string
              country:
                type: string
              disambiguated-id:
                type: string
              disambiguated-source:
                type: string
              start-date:
                type: string
              end-date:
                type: string
              processed-at:
                type: string
                format: date-time
              status:
                type: string
              orcid:
                type: string
                format: "^[0-9]{4}-?[0-9]{4}-?[0-9]{4}-?[0-9]{4}$"
                description: "User ORCID ID"
        """
        login_user(request.oauth.user)
        if request.content_type in ["text/csv", "text/tsv"]:
            task = PropertyRecord.load_from_csv(request.data.decode("utf-8"), filename=self.filename)
            enqueue_task_records(task)
            return self.jsonify_task(task)
        return self.handle_task()


class PropertyAPI(PropertyListAPI):
    """Property task services."""

    def get(self, task_id):
        """
        Retrieve the specified property task.

        ---
        tags:
          - "properties"
        summary: "Retrieve the specified property task."
        description: "Retrieve the specified property task."
        produces:
          - "application/json"
        parameters:
          - name: "task_id"
            required: true
            in: "path"
            description: "Property task ID."
            type: "integer"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PropertyTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.jsonify_task(task_id)

    def post(self, task_id):
        """Upload the task and completely override the property task.

        ---
        tags:
          - "properties"
        summary: "Update the property task."
        description: "Update the property task."
        consumes:
          - application/json
          - text/yaml
        definitions:
        parameters:
          - name: "task_id"
            in: "path"
            description: "Property task ID."
            required: true
            type: "integer"
          - in: body
            name: propertyTask
            description: "Property task."
            schema:
              $ref: "#/definitions/PropertyTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PropertyTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def put(self, task_id):
        """Update the property task.

        ---
        tags:
          - "properties"
        summary: "Update the property task."
        description: "Update the property task."
        consumes:
          - application/json
          - text/yaml
        parameters:
          - name: "task_id"
            in: "path"
            description: "Property task ID."
            required: true
            type: "integer"
          - in: body
            name: propertyTask
            description: "Property task."
            schema:
              $ref: "#/definitions/PropertyTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PropertyTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def patch(self, task_id):
        """Update the property task.

        ---
        tags:
          - "properties"
        summary: "Update the property task."
        description: "Update the property task."
        consumes:
          - application/json
          - text/yaml
        parameters:
          - name: "task_id"
            in: "path"
            description: "Property task ID."
            required: true
            type: "integer"
          - in: body
            name: propertyTask
            description: "Property task."
            schema:
              $ref: "#/definitions/PropertyTask"
        produces:
          - "application/json"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/PropertyTask"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.handle_task(task_id)

    def delete(self, task_id):
        """Delete the specified property task.

        ---
        tags:
          - "properties"
        summary: "Delete the specified property task."
        description: "Delete the specified property task."
        parameters:
          - name: "task_id"
            in: "path"
            description: "Property task ID."
            required: true
            type: "integer"
        produces:
          - "application/json"
        responses:
          200:
            description: "Successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        return self.delete_task(task_id)


api.add_resource(PropertyListAPI, "/api/v1.0/properties")
api.add_resource(PropertyAPI, "/api/v1.0/properties/<int:task_id>")


class UserListAPI(AppResourceList):
    """User list data service."""

    def get(self):
        """
        Return the list of the user belonging to the organisation.

        ---
        definitions:
        - schema:
            id: HubUser
            properties:
              orcid:
                type: "string"
                format: "^[0-9]{4}-?[0-9]{4}-?[0-9]{4}-?[0-9]{4}$"
                description: "User ORCID ID"
              email:
                type: "string"
              eppn:
                type: "string"
              confirmed:
                type: "boolean"
              updated-at:
                type: "string"
                format: date-time
              name:
                type: "string"
            required:
            - email
        tags:
          - "users"
        summary: "Retrieve the list of all users."
        description: "Retrieve the list of all users."
        produces:
          - "application/json"
        parameters:
          - in: query
            name: from_date
            type: string
            minimum: 0
            description: The date starting from which user accounts were created and/or updated.
          - in: query
            name: to_date
            type: string
            minimum: 0
            description: The date until which user accounts were created and/or updated.
          - in: query
            name: page
            type: integer
            minimum: 0
            default: 1
            description: The number of the page of retrieved data starting counting from 1
          - in: query
            name: page_size
            type: integer
            minimum: 0
            default: 20
            description: The size of the data page
        responses:
          200:
            description: "successful operation"
            schema:
              id: UserListApiResponse
              type: array
              items:
                $ref: "#/definitions/HubUser"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        login_user(request.oauth.user)
        users = User.select().where(User.organisation == current_user.organisation)
        for a in ["from_date", "to_date"]:
            v = request.args.get(a)
            if not v:
                continue
            try:
                v = datetime.strptime(v, "%Y-%m-%d")
            except ValueError as ex:
                return jsonify({
                    "error": f"Failed to parse the date for \"{a}\": {v}",
                    "message": str(ex)
                }), 422
            if a.startswith("from"):
                users = users.where((User.created_at >= v) | (User.updated_at >= v))
            else:
                users = users.where((User.created_at <= v) & (User.updated_at <= v))
        return self.api_response(
            users,
            only=[User.email, User.eppn, User.name, User.orcid, User.confirmed, User.updated_at])

    def post(self):
        """
        Create a bare user account that could be used for further ORCID profile management.

        ---
        tags:
          - "users"
        summary: "Create a bare user account."
        description: "Create a bare user account that could be used for further ORCID profile management."
        produces:
          - "application/json"
        consumes:
          - "application/json"
        parameters:
          - name: "body"
            in: "body"
            description: "User data."
            required: true
            schema:
              $ref: "#/definitions/HubUser"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/HubUser"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        return self.handle_user()


api.add_resource(UserListAPI, "/api/v1.0/users")


class UserAPI(AppResource):
    """User data service."""

    def get(self, identifier=None):
        """
        Verify if a user with given email address or ORCID ID exists.

        ---
        tags:
          - "users"
        summary: "Get user by user email or ORCID ID"
        description: ""
        produces:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "The unique identifier (email, ORCID iD, or eppn)"
            required: true
            type: "string"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/HubUser"
          400:
            description: "Invalid identifier supplied"
          404:
            description: "User not found"
        """
        identifier = identifier.strip()
        if validators.email(identifier):
            user = User.select().where((User.email == identifier)
                                       | (User.eppn == identifier)).first()
        elif ORCID_ID_REGEX.match(identifier):
            try:
                models.validate_orcid_id(identifier)
            except Exception as ex:
                return jsonify({"error": f"Incorrect identifier value '{identifier}': {ex}"}), 400
            user = User.select().where(User.orcid == identifier).first()
        else:
            return jsonify({"error": f"Incorrect identifier value: {identifier}."}), 400
        if user is None or (user.organisation != request.oauth.client.org
                            and not UserOrg.select().where(
                                UserOrg.org == request.oauth.client.org,
                                UserOrg.user == user,
                            ).exists()):
            return jsonify({
                "error": f"User with specified identifier '{identifier}' not found."
            }), 404

        return jsonify({
            "confirmed": user.confirmed,
            "email": user.email,
            "eppn": user.eppn,
            "orcid": user.orcid,
            "updated-at": user.updated_at,
        }), 200

    def post(self, identifier=None):
        """
        Update a user account.

        ---
        tags:
          - "users"
        summary: "Update a user account."
        description: "Update a user account data."
        produces:
          - "application/json"
        consumes:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
          - name: "body"
            required: true
            in: "body"
            description: "User data."
            schema:
              $ref: "#/definitions/HubUser"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/HubUser"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        return self.handle_user(identifier)

    def put(self, identifier=None):
        """
        Update a user account.

        ---
        tags:
          - "users"
        summary: "Update a user account."
        description: "Update a user account data."
        produces:
          - "application/json"
        consumes:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
          - name: "body"
            required: true
            in: "body"
            description: "User data."
            schema:
              $ref: "#/definitions/HubUser"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/HubUser"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        return self.handle_user(identifier)

    def patch(self, identifier=None):
        """
        Update a user account.

        ---
        tags:
          - "users"
        summary: "Update a user account."
        description: "Update a user account data."
        produces:
          - "application/json"
        consumes:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
          - name: "body"
            required: true
            in: "body"
            description: "User data."
            schema:
              $ref: "#/definitions/HubUser"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/HubUser"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        return self.handle_user(identifier)

    def delete(self, identifier=None):
        """
        Delete a user account.

        ---
        tags:
          - "users"
        summary: "Delete a user account."
        description: "Delete a user account data."
        consumes:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
        responses:
          202:
            description: "successful operation"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
          422:
            description: "Unprocessable Entity"
        """
        return self.handle_user(identifier)


api.add_resource(UserAPI, "/api/v1.0/users/<identifier>")


class TokenAPI(MethodView):
    """ORCID access token service."""

    decorators = [
        oauth.require_oauth(),
    ]

    def get_user(self, identifier):
        """Get the user account record. Returns a tuple: (user, error_message, status_code)."""
        identifier = identifier.strip()
        if validators.email(identifier):
            user = User.select().where((User.email == identifier)
                                       | (User.eppn == identifier)).first()
        elif ORCID_ID_REGEX.match(identifier):
            try:
                models.validate_orcid_id(identifier)
            except Exception as ex:
                return None, f"Incorrect identifier value '{identifier}': {ex}", 400
            user = User.select().where(User.orcid == identifier).first()
        else:
            return None, f"Incorrect identifier value: {identifier}.", 400

        org = request.oauth.client.org
        if user is None or (user.organisation != org
                            and not UserOrg.select().where(
                                UserOrg.org == org,
                                UserOrg.user == user,
                            ).exists()):
            return None, f"User with specified identifier '{identifier}' not found.", 404
        return user, None, None

    def get(self, identifier=None):
        """
        Retrieve user ORCID API access token and refresh tokens.

        ---
        definitions:
        - schema:
            id: OrcidToken
            properties:
              access_token:
                type: "string"
                description: "ORCID API user profile access token"
              refresh_token:
                type: "string"
                description: "ORCID API user profile refresh token"
              scopes:
                type: "string"
                description: "ORCID API user token scopes"
              issue_time:
                type: "string"
              expires_in:
                type: "integer"
        tags:
          - "token"
        summary: "Retrieves user ORCID API access and refresh tokens."
        description: ""
        produces:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
        responses:
          200:
            description: "successful operation"
            schema:
              type: array
              items:
                $ref: "#/definitions/OrcidToken"
          400:
            description: "Invalid identifier supplied"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        user, error_message, status_code = self.get_user(identifier)
        if not user:
            return jsonify(error=error_message), status_code

        org = request.oauth.client.org
        tokens = OrcidToken.select().where(OrcidToken.user == user,
                                           OrcidToken.org == org)

        return jsonify([
            t.to_dict(recurse=False,
                      only=[
                          OrcidToken.access_token, OrcidToken.refresh_token, OrcidToken.scopes,
                          OrcidToken.issue_time, OrcidToken.expires_in
                      ]) for t in tokens
        ]), 200

    def post(self, identifier=None):
        """
        Add an ORCID API access token for a user account.

        ---
        tags:
          - "token"
        summary: "Add an access token for a user specified by a unique identifier."
        description: "Add an access token for a user specified by a unique identifier (email or orcid)."
        produces:
          - "application/json"
        parameters:
          - name: "identifier"
            in: "path"
            description: "User identifier (either email, eppn or ORCID ID)"
            required: true
            type: "string"
          - name: "body"
            in: "body"
            required: true
            description: "ORCID API access token."
            schema:
              $ref: "#/definitions/OrcidToken"
        responses:
          200:
            description: "successful operation"
            schema:
              $ref: "#/definitions/OrcidToken"
          400:
            description: "Invalid identifier supplied"
          401:
            $ref: "#/responses/Unauthorized"
          403:
            $ref: "#/responses/AccessDenied"
          404:
            $ref: "#/responses/NotFound"
        """
        user, error_message, status_code = self.get_user(identifier)
        if not user:
            return jsonify(error=error_message), status_code

        data = request.get_json()
        access_token = data.get("access_token")
        if not validate_uuid4(access_token):
            return jsonify({
                "error":
                f"Invalid access token value: {access_token}."
            }), 400

        refresh_token = data.get("refresh_token")
        if refresh_token and not validate_uuid4(refresh_token):
            return jsonify({
                "error":
                f"Invalid refresh token value: {access_token}."
            }), 400

        scopes = data.get("scopes")
        if not SCOPE_REGEX.match(scopes):
            return jsonify({
                "error":
                f"Invalid scope value: {scopes}."
            }), 400

        issue_time = data.get("issue_time")
        if issue_time:
            try:
                issue_time = dateutil.parser.parse(issue_time)
            except ValueError as ex:
                return jsonify({
                    "error":
                    f"Invalid access token issue time value: {issue_time}: {ex.args}."
                }), 400

        expires_in = data.get("expires_in")
        if expires_in:
            try:
                expires_in = int(expires_in)
            except:
                return jsonify({
                    "error":
                    f"Invalid access token expires in value: {expires_in}."
                }), 400

        org = request.oauth.client.org
        q = OrcidToken.select().where(OrcidToken.user == user, OrcidToken.org == org)
        if scopes:
            q = q.where(OrcidToken.scopes == scopes)

        if q.exists():
            return jsonify({
                "error":
                f"Token for the users {user} ({identifier}) with the scope '{scopes}' already exists."
            }), 400

        try:
            token = OrcidToken.create(user=user,
                                      org=org,
                                      scopes=scopes,
                                      issue_time=issue_time,
                                      expires_in=expires_in,
                                      access_token=access_token,
                                      refresh_token=refresh_token)
        except OrcidToken.DoesNotExist:
            return jsonify({
                "error":
                f"Token for the users {user} ({identifier}) affiliated with {org} not found."
            }), 404

        return jsonify(
            token.to_dict(recurse=False,
                          only=[
                              OrcidToken.access_token, OrcidToken.refresh_token, OrcidToken.scopes,
                              OrcidToken.issue_time, OrcidToken.expires_in
                          ])), 201


app.add_url_rule("/api/v1.0/tokens/<identifier>", view_func=TokenAPI.as_view("tokens"))


def get_spec(app):
    """Build API swagger scecifiction."""
    swag = swagger(app)
    swag["info"]["version"] = "1.0"
    swag["info"]["title"] = "ORCID HUB API"
    # swag["basePath"] = "/api/v1.0"
    swag["host"] = request.host  # "dev.orcidhub.org.nz"
    swag["consumes"] = ["application/json"]
    swag["produces"] = ["application/json"]
    swag["schemes"] = [request.scheme]
    swag["securityDefinitions"] = {
        "application": {
            "type": "oauth2",
            "tokenUrl": url_for("access_token", _external=True),
            "flow": "application",
            "scopes": {
                "write": "allows modifying resources",
                "read": "allows reading resources",
            }
        }
    }
    swag["security"] = [
        {
            "application": [
                "read",
                "write",
            ]
        },
    ]
    swag["tags"] = [
        {
            "name": "affiliations",
            "description": "Affiliation data management APIs",
            "externalDocs": {
                "url": "http://docs.orcidhub.org.nz/en/latest/writing_affiliation_items.html"
            },
        },
        {
            "name": "funds",
            "description": "Fund data management APIs",
            "externalDocs": {
                "url": "http://docs.orcidhub.org.nz/en/latest/writing_funding_items.html"
            },
        },
        {
            "name": "orcid-proxy",
            "description": "ORCID API proxy",
            "externalDocs": {
                "url": "https://api.sandbox.orcid.org"
            },
        },
        {
            "name": "webhooks",
            "description": "ORCID Webhook management",
            "externalDocs": {
                "url": "https://members.orcid.org/api/tutorial/webhooks"
            },
        },
    ]
    swag["parameters"] = {
        "orcidParam": {
            "in": "path",
            "name": "orcid",
            "required": True,
            "type": "string",
            "description": "User ORCID ID.",
            "format": "^[0-9]{4}-?[0-9]{4}-?[0-9]{4}-?[0-9]{4}$",
        },
        "versionParam": {
            "in": "path",
            "name": "version",
            "required": True,
            "description": "ORCID API version",
            "type": "string",
            "enum": [
                "v2.0",
                "v2.1",
                "v3.0_rc1",
                "v3.0_rc2s",
                "v3.0",
            ],
        },
        "pathParam": {
            "in": "path",
            "name": "path",
            "required": True,
            "type": "string",
            "description": "The rest of the ORCID API entry point URL.",
        },
    }
    # Common responses:
    swag["responses"] = {
            "AccessDenied": {
                "description": "Access Denied",
                "schema": {"$ref": "#/definitions/Error"}
            },
            "Unauthorized": {
                "description": "Unauthorized",
                "schema": {"$ref": "#/definitions/Error"}
            },
            "NotFound": {
                "description": "The specified resource was not found",
                "schema": {"$ref": "#/definitions/Error"}
            },
    }
    swag["definitions"]["Error"] = {
        "properties": {
            "error": {
                "type": "string",
                "description": "Error type/name."
            },
            "message": {
                "type": "string",
                "description": "Error details explaining message."
            },
        }
    }
    # Webhooks:
    put_responses = {
        "201": {
            "description": "A webhoook successfully set up.",
        },
        "415": {
            "description": "Invalid call-back URL or missing ORCID iD.",
            "schema": {
                "$ref": "#/definitions/Error"
            },
        },
        "404": {
            "description": "Invalid ORCID iD.",
            "schema": {
                "$ref": "#/definitions/Error"
            },
        },
    }
    delete_responses = {
        "204": {
            "description": "A webhoook successfully unregistered.",
        },
        "415": {
            "description": "Invalid call-back URL or missing ORCID iD.",
            "schema": {
                "$ref": "#/definitions/Error"
            },
        },
        "404": {
            "description": "Invalid ORCID iD.",
            "schema": {
                "$ref": "#/definitions/Error"
            },
        },
    }
    swag["paths"]["/api/v1.0/{orcid}/webhook"] = {
        "parameters": [swag["parameters"]["orcidParam"]],
        "put": {
            "tags": ["webhooks"],
            "responses": put_responses,
        },
        "delete": {
            "tags": ["webhooks"],
            "responses": delete_responses,
        }
    }
    swag["paths"]["/api/v1.0/{orcid}/webhook/{callback_url}"] = {
        "parameters": [
            swag["parameters"]["orcidParam"],
            {
                "in": "path",
                "name": "callback_url",
                "required": True,
                "type": "string",
                "description":
                "The call-back URL that will receive a POST request when an update of a ORCID profile occurs.",
            },
        ],
        "put": {
            "tags": ["webhooks"],
            "responses": put_responses,
        },
        "delete": {
            "tags": ["webhooks"],
            "responses": delete_responses,
        }
    }
    # Proxy:
    swag["paths"]["/orcid/api/{version}/{orcid}"] = {
        "parameters": [
            swag["parameters"]["versionParam"],
            swag["parameters"]["orcidParam"],
        ],
        "get": {
            "tags": ["orcid-proxy"],
            "produces": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "responses": {
                "200": {
                    "description": "Successful operation",
                    "schema": {
                        "type": "object"
                    }
                },
                "403": {
                    "description": "The user hasn't granted acceess to the profile.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "404": {
                    "description": "Resource not found",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "415": {
                    "description": "Missing or invalid ORCID iD.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
            },
        },
    }
    swag["paths"]["/orcid/api/{version}/{orcid}/{path}"] = {
        "parameters": [
            swag["parameters"]["versionParam"],
            swag["parameters"]["orcidParam"],
            swag["parameters"]["pathParam"],
        ],
        "delete": {
            "tags": ["orcid-proxy"],
            "produces": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "responses": {
                "204": {
                    "description": "Record deleted",
                    "schema": {
                        "type": "object"
                    }
                },
                "403": {
                    "description": "The user hasn't granted acceess to the profile.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "404": {
                    "description": "Resource not found",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "415": {
                    "description": "Missing or invalid ORCID iD.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
            },
        },
        "get": {
            "tags": ["orcid-proxy"],
            "produces": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "responses": {
                "200": {
                    "description": "Successful operation",
                    "schema": {
                        "type": "object"
                    }
                },
                "403": {
                    "description": "The user hasn't granted acceess to the profile.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "404": {
                    "description": "Resource not found",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "415": {
                    "description": "Missing or invalid ORCID iD.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
            },
        },
        "post": {
            "tags": ["orcid-proxy"],
            "consumes": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "produces": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "parameters": [{
                "in": "body",
                "name": "body",
                "required": False,
                "schema": {
                    "type": "object"
                }
            }],
            "responses": {
                "200": {
                    "description": "successful operation",
                    "schema": {
                        "type": "object"
                    }
                },
                "403": {
                    "description": "The user hasn't granted acceess to the profile.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "404": {
                    "description": "Resource not found",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "415": {
                    "description": "Missing or invalid ORCID iD.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
            },
        },
        "put": {
            "tags": ["orcid-proxy"],
            "consumes": [
                "application/vnd.orcid+xml; qs=5", "application/orcid+xml; qs=3",
                "application/xml", "application/vnd.orcid+json; qs=4",
                "application/orcid+json; qs=2", "application/json"
            ],
            "parameters": [{
                "in": "body",
                "name": "body",
                "required": False,
                "schema": {
                    "type": "object"
                }
            }],
            "responses": {
                "200": {
                    "description": "successful operation",
                    "schema": {
                        "type": "object"
                    }
                },
                "403": {
                    "description": "The user hasn't granted acceess to the profile.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "404": {
                    "description": "Resource not found",
                    "schema": {"$ref": "#/definitions/Error"}
                },
                "415": {
                    "description": "Missing or invalid ORCID iD.",
                    "schema": {"$ref": "#/definitions/Error"}
                },
            }
        }
    }
    return swag


@app.route("/spec.json")
@app.route("/spec.yml")
@app.route("/spec.yaml")
@app.route("/spec")
def spec():
    """Return the specification of the API."""
    swag = get_spec(app)
    best = request.accept_mimetypes.best_match(["text/yaml", "application/x-yaml"])
    path = request.path
    if path.endswith(".yaml") or path.endswith(".yml") or (best in (
            "text/yaml",
            "application/x-yaml",
    ) and request.accept_mimetypes[best] > request.accept_mimetypes["application/json"]):
        return yamlfy(swag)
    else:
        return jsonify(swag)


@app.route("/api-docs")
@roles_required(Role.TECHNICAL, Role.SUPERUSER)
def api_docs():
    """Show Swagger UI for the latest/current Hub API and Data API."""
    client = Client.select().where(Client.org == current_user.organisation).first()
    return render_template("swaggerui.html", client=client)


def yamlfy(*args, **kwargs):
    """Create respose in YAML just like jsonify does it for JSON."""
    if args and kwargs:
        raise TypeError('yamlfy() behavior undefined when passed both args and kwargs')
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    else:
        data = args or kwargs

    return current_app.response_class((dump_yaml(data), '\n'), mimetype="text/yaml")


@app.route("/orcid/api/<string:version>/<string:orcid>", methods=["GET", "POST", "PUT", "DELETE"])
@app.route(
    "/orcid/api/<string:version>/<string:orcid>/<path:rest>",
    methods=["GET", "POST", "PUT", "DELETE"])
@oauth.require_oauth()
def orcid_proxy(version, orcid, rest=None):
    """Handle proxied request..."""
    login_user(request.oauth.user)
    if not ORCID_API_VERSION_REGEX.match(version):
        return jsonify({
            "error": "Resource not found",
            "message": f"Incorrect version: {version}"
        }), 404
    try:
        validate_orcid_id(orcid)
    except Exception as ex:
        return jsonify({"error": str(ex), "message": "Missing or invalid ORCID iD."}), 415
    token = OrcidToken.select().join(User).where(
        User.orcid == orcid, OrcidToken.org == current_user.organisation).first()
    if not token:
        return jsonify({"message": "The user hasn't granted access to the user profile"}), 403

    orcid_api_host_url = app.config["ORCID_API_HOST_URL"]
    # CHUNK_SIZE = 1024
    # TODO: sanitize headers
    headers = {
        h: v
        for h, v in request.headers if h in
        ["Cache-Control", "User-Agent", "Accept", "Accept-Encoding", "Connection", "Content-Type"]
    }
    headers["Authorization"] = f"Bearer {token.access_token}"
    url = f"{orcid_api_host_url}{version}/{orcid}"
    # Swagger-UI sets 'path' to 'undefined':
    if rest and rest != "undefined":
        url += '/' + rest

    proxy_req = requests.Request(
        request.method, url, data=request.stream, headers=headers).prepare()
    session = requests.Session()
    # TODO: add time-out
    resp = session.send(proxy_req, stream=True)

    def generate():
        # for chunk in resp.raw.stream(decode_content=False, amt=CHUNK_SIZE):
        for chunk in resp.raw.stream(decode_content=False):
            yield chunk

    # TODO: verify if flask can create chunked responses: Transfer-Encoding: chunked
    proxy_headers = [(h, v) for h, v in resp.raw.headers.items() if h not in [
        "Transfer-Encoding",
    ]]
    proxy_resp = Response(
        stream_with_context(generate()), headers=proxy_headers, status=resp.status_code)
    return proxy_resp


@app.route("/api/v1.0/<string:orcid>/webhook", methods=["PUT", "DELETE"])
@app.route("/api/v1.0/<string:orcid>/webhook/<path:callback_url>", methods=["PUT", "DELETE"])
@oauth.require_oauth()
def register_webhook(orcid, callback_url=None):
    """Handle webhook registration for an individual user with direct client call-back."""
    login_user(request.oauth.user)

    try:
        validate_orcid_id(orcid)
    except Exception as ex:
        return jsonify({"error": "Missing or invalid ORCID iD.", "message": str(ex)}), 415
    if callback_url == "undefined":
        callback_url = None
    if callback_url:
        callback_url = unquote(callback_url)
        if not is_valid_url(callback_url):
            return jsonify({
                "error": "Invalid call-back URL",
                "message": f"Invalid call-back URL: {callback_url}"
            }), 415

    try:
        user = User.get(orcid=orcid)
    except User.DoesNotExist:
        return jsonify({
            "error": "Invalid ORCID iD.",
            "message": f"User with given ORCID ID '{orcid}' doesn't exist."
        }), 404

    orcid_resp = register_orcid_webhook(user, callback_url, delete=request.method == "DELETE")
    resp = make_response('', orcid_resp.status_code if orcid_resp else 204)
    if orcid_resp and "Location" in orcid_resp.headers:
        resp.headers["Location"] = orcid_resp.headers["Location"]

    return resp



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.addresses_v30_rc1 import AddressesV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.biography_v30_rc1 import BiographyV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.emails_v30_rc1 import EmailsV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.keywords_v30_rc1 import KeywordsV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.last_modified_date_v30_rc1 import LastModifiedDateV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.name_v30_rc1 import NameV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.other_names_v30_rc1 import OtherNamesV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.person_external_identifiers_v30_rc1 import PersonExternalIdentifiersV30Rc1  # noqa: F401,E501
from orcid_api_v3.models.researcher_urls_v30_rc1 import ResearcherUrlsV30Rc1  # noqa: F401,E501


class PersonV30Rc1(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'last_modified_date': 'LastModifiedDateV30Rc1',
        'name': 'NameV30Rc1',
        'other_names': 'OtherNamesV30Rc1',
        'biography': 'BiographyV30Rc1',
        'researcher_urls': 'ResearcherUrlsV30Rc1',
        'emails': 'EmailsV30Rc1',
        'addresses': 'AddressesV30Rc1',
        'keywords': 'KeywordsV30Rc1',
        'external_identifiers': 'PersonExternalIdentifiersV30Rc1',
        'path': 'str'
    }

    attribute_map = {
        'last_modified_date': 'last-modified-date',
        'name': 'name',
        'other_names': 'other-names',
        'biography': 'biography',
        'researcher_urls': 'researcher-urls',
        'emails': 'emails',
        'addresses': 'addresses',
        'keywords': 'keywords',
        'external_identifiers': 'external-identifiers',
        'path': 'path'
    }

    def __init__(self, last_modified_date=None, name=None, other_names=None, biography=None, researcher_urls=None, emails=None, addresses=None, keywords=None, external_identifiers=None, path=None):  # noqa: E501
        """PersonV30Rc1 - a model defined in Swagger"""  # noqa: E501
        self._last_modified_date = None
        self._name = None
        self._other_names = None
        self._biography = None
        self._researcher_urls = None
        self._emails = None
        self._addresses = None
        self._keywords = None
        self._external_identifiers = None
        self._path = None
        self.discriminator = None
        if last_modified_date is not None:
            self.last_modified_date = last_modified_date
        if name is not None:
            self.name = name
        if other_names is not None:
            self.other_names = other_names
        if biography is not None:
            self.biography = biography
        if researcher_urls is not None:
            self.researcher_urls = researcher_urls
        if emails is not None:
            self.emails = emails
        if addresses is not None:
            self.addresses = addresses
        if keywords is not None:
            self.keywords = keywords
        if external_identifiers is not None:
            self.external_identifiers = external_identifiers
        if path is not None:
            self.path = path

    @property
    def last_modified_date(self):
        """Gets the last_modified_date of this PersonV30Rc1.  # noqa: E501


        :return: The last_modified_date of this PersonV30Rc1.  # noqa: E501
        :rtype: LastModifiedDateV30Rc1
        """
        return self._last_modified_date

    @last_modified_date.setter
    def last_modified_date(self, last_modified_date):
        """Sets the last_modified_date of this PersonV30Rc1.


        :param last_modified_date: The last_modified_date of this PersonV30Rc1.  # noqa: E501
        :type: LastModifiedDateV30Rc1
        """

        self._last_modified_date = last_modified_date

    @property
    def name(self):
        """Gets the name of this PersonV30Rc1.  # noqa: E501


        :return: The name of this PersonV30Rc1.  # noqa: E501
        :rtype: NameV30Rc1
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this PersonV30Rc1.


        :param name: The name of this PersonV30Rc1.  # noqa: E501
        :type: NameV30Rc1
        """

        self._name = name

    @property
    def other_names(self):
        """Gets the other_names of this PersonV30Rc1.  # noqa: E501


        :return: The other_names of this PersonV30Rc1.  # noqa: E501
        :rtype: OtherNamesV30Rc1
        """
        return self._other_names

    @other_names.setter
    def other_names(self, other_names):
        """Sets the other_names of this PersonV30Rc1.


        :param other_names: The other_names of this PersonV30Rc1.  # noqa: E501
        :type: OtherNamesV30Rc1
        """

        self._other_names = other_names

    @property
    def biography(self):
        """Gets the biography of this PersonV30Rc1.  # noqa: E501


        :return: The biography of this PersonV30Rc1.  # noqa: E501
        :rtype: BiographyV30Rc1
        """
        return self._biography

    @biography.setter
    def biography(self, biography):
        """Sets the biography of this PersonV30Rc1.


        :param biography: The biography of this PersonV30Rc1.  # noqa: E501
        :type: BiographyV30Rc1
        """

        self._biography = biography

    @property
    def researcher_urls(self):
        """Gets the researcher_urls of this PersonV30Rc1.  # noqa: E501


        :return: The researcher_urls of this PersonV30Rc1.  # noqa: E501
        :rtype: ResearcherUrlsV30Rc1
        """
        return self._researcher_urls

    @researcher_urls.setter
    def researcher_urls(self, researcher_urls):
        """Sets the researcher_urls of this PersonV30Rc1.


        :param researcher_urls: The researcher_urls of this PersonV30Rc1.  # noqa: E501
        :type: ResearcherUrlsV30Rc1
        """

        self._researcher_urls = researcher_urls

    @property
    def emails(self):
        """Gets the emails of this PersonV30Rc1.  # noqa: E501


        :return: The emails of this PersonV30Rc1.  # noqa: E501
        :rtype: EmailsV30Rc1
        """
        return self._emails

    @emails.setter
    def emails(self, emails):
        """Sets the emails of this PersonV30Rc1.


        :param emails: The emails of this PersonV30Rc1.  # noqa: E501
        :type: EmailsV30Rc1
        """

        self._emails = emails

    @property
    def addresses(self):
        """Gets the addresses of this PersonV30Rc1.  # noqa: E501


        :return: The addresses of this PersonV30Rc1.  # noqa: E501
        :rtype: AddressesV30Rc1
        """
        return self._addresses

    @addresses.setter
    def addresses(self, addresses):
        """Sets the addresses of this PersonV30Rc1.


        :param addresses: The addresses of this PersonV30Rc1.  # noqa: E501
        :type: AddressesV30Rc1
        """

        self._addresses = addresses

    @property
    def keywords(self):
        """Gets the keywords of this PersonV30Rc1.  # noqa: E501


        :return: The keywords of this PersonV30Rc1.  # noqa: E501
        :rtype: KeywordsV30Rc1
        """
        return self._keywords

    @keywords.setter
    def keywords(self, keywords):
        """Sets the keywords of this PersonV30Rc1.


        :param keywords: The keywords of this PersonV30Rc1.  # noqa: E501
        :type: KeywordsV30Rc1
        """

        self._keywords = keywords

    @property
    def external_identifiers(self):
        """Gets the external_identifiers of this PersonV30Rc1.  # noqa: E501


        :return: The external_identifiers of this PersonV30Rc1.  # noqa: E501
        :rtype: PersonExternalIdentifiersV30Rc1
        """
        return self._external_identifiers

    @external_identifiers.setter
    def external_identifiers(self, external_identifiers):
        """Sets the external_identifiers of this PersonV30Rc1.


        :param external_identifiers: The external_identifiers of this PersonV30Rc1.  # noqa: E501
        :type: PersonExternalIdentifiersV30Rc1
        """

        self._external_identifiers = external_identifiers

    @property
    def path(self):
        """Gets the path of this PersonV30Rc1.  # noqa: E501


        :return: The path of this PersonV30Rc1.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this PersonV30Rc1.


        :param path: The path of this PersonV30Rc1.  # noqa: E501
        :type: str
        """

        self._path = path

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(PersonV30Rc1, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PersonV30Rc1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



#!/usr/bin/env python
"""
Very simple HTTP server in python.

Usage::
    ./dummy-web-server.py [<port>]

Send a GET request::
    curl http://localhost

Send a HEAD request::
    curl -I http://localhost

Send a POST request::
    curl -d "foo=bar&bin=baz" http://localhost

"""
from http.server import BaseHTTPRequestHandler, HTTPServer

import logging
try:
    import coloredlogs
    coloredlogs.install()
except ImportError:
    pass

host_name = ""
host_port = 8080


class UpdateHandler(BaseHTTPRequestHandler):
    """Webhook update even handler."""

    def do_GET(self):  # noqa: N802
        """Handle GET request."""
        self.send_response(204)
        self.end_headers()
        self.wfile.write(b'')

    def do_POST(self):  # noqa: N802
        """Handle POST request."""
        if "Content-Length" in self.headers:
            content_length = int(self.headers['Content-Length'])
            post_data = self.rfile.read(content_length)
            if post_data:
                logging.info(f"Received: {post_data.decode()}")
        self.log_message("*** Headers:\n%s", self.headers)
        self.log_message("Source: %s", self.headers.get("X-Forwarded-For"))

        self.send_response(204)
        self.end_headers()
        self.wfile.write(b'')

    def log_message(self, format, *args):
        """Log a message."""
        logging.info(format, *args)

    def log_error(self, format, *args):
        """Log an error message."""
        logging.error(format, *args)


server = HTTPServer((host_name, host_port), UpdateHandler)
logging.info(f"Server Starts {host_name}, port: {host_port}")

try:
    server.serve_forever()
except KeyboardInterrupt:
    pass

server.server_close()
logging.info(f"Server Stops {host_name}, port: {host_port}")



# -*- coding: utf-8 -*-
"""Various utilities."""

from . import utils as tasks, rq
from datetime import datetime


def setup():
    """Set up the application schedule. Remove any already scheduled jobs."""
    scheduler = rq.get_scheduler()
    for job in scheduler.get_jobs():
        job.delete()

    # NB! add result_ttl! Otherwise it won't get rescheduled
    tasks.process_tasks.schedule(datetime.utcnow(), interval=3600, result_ttl=-1, job_id="*PROCESS-TASKS*")
    tasks.send_orcid_update_summary.cron("0 0 1 * *", "*ORCID-UPDATE-SUMMARY*")



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.day_v20 import DayV20  # noqa: F401,E501
from orcid_api_v3.models.month_v20 import MonthV20  # noqa: F401,E501
from orcid_api_v3.models.year_v20 import YearV20  # noqa: F401,E501


class FuzzyDateV20(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'year': 'YearV20',
        'month': 'MonthV20',
        'day': 'DayV20'
    }

    attribute_map = {
        'year': 'year',
        'month': 'month',
        'day': 'day'
    }

    def __init__(self, year=None, month=None, day=None):  # noqa: E501
        """FuzzyDateV20 - a model defined in Swagger"""  # noqa: E501
        self._year = None
        self._month = None
        self._day = None
        self.discriminator = None
        self.year = year
        if month is not None:
            self.month = month
        if day is not None:
            self.day = day

    @property
    def year(self):
        """Gets the year of this FuzzyDateV20.  # noqa: E501


        :return: The year of this FuzzyDateV20.  # noqa: E501
        :rtype: YearV20
        """
        return self._year

    @year.setter
    def year(self, year):
        """Sets the year of this FuzzyDateV20.


        :param year: The year of this FuzzyDateV20.  # noqa: E501
        :type: YearV20
        """
        if year is None:
            raise ValueError("Invalid value for `year`, must not be `None`")  # noqa: E501

        self._year = year

    @property
    def month(self):
        """Gets the month of this FuzzyDateV20.  # noqa: E501


        :return: The month of this FuzzyDateV20.  # noqa: E501
        :rtype: MonthV20
        """
        return self._month

    @month.setter
    def month(self, month):
        """Sets the month of this FuzzyDateV20.


        :param month: The month of this FuzzyDateV20.  # noqa: E501
        :type: MonthV20
        """

        self._month = month

    @property
    def day(self):
        """Gets the day of this FuzzyDateV20.  # noqa: E501


        :return: The day of this FuzzyDateV20.  # noqa: E501
        :rtype: DayV20
        """
        return self._day

    @day.setter
    def day(self, day):
        """Sets the day of this FuzzyDateV20.


        :param day: The day of this FuzzyDateV20.  # noqa: E501
        :type: DayV20
        """

        self._day = day

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(FuzzyDateV20, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, FuzzyDateV20):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    ORCID Member

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: Latest

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six
from orcid_api_v3.models.contributor_email_v20 import ContributorEmailV20  # noqa: F401,E501
from orcid_api_v3.models.contributor_orcid_v20 import ContributorOrcidV20  # noqa: F401,E501
# from orcid_api_v3.models.contributor_v20 import ContributorV20  # noqa: F401,E501
from orcid_api_v3.models.credit_name_v20 import CreditNameV20  # noqa: F401,E501


class ContributorV20(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'contributor_orcid': 'ContributorOrcidV20',
        'credit_name': 'CreditNameV20',
        'contributor_email': 'ContributorEmailV20',
        'contributor_attributes': 'ContributorV20'
    }

    attribute_map = {
        'contributor_orcid': 'contributor-orcid',
        'credit_name': 'credit-name',
        'contributor_email': 'contributor-email',
        'contributor_attributes': 'contributor-attributes'
    }

    def __init__(self, contributor_orcid=None, credit_name=None, contributor_email=None, contributor_attributes=None):  # noqa: E501
        """ContributorV20 - a model defined in Swagger"""  # noqa: E501
        self._contributor_orcid = None
        self._credit_name = None
        self._contributor_email = None
        self._contributor_attributes = None
        self.discriminator = None
        if contributor_orcid is not None:
            self.contributor_orcid = contributor_orcid
        if credit_name is not None:
            self.credit_name = credit_name
        if contributor_email is not None:
            self.contributor_email = contributor_email
        if contributor_attributes is not None:
            self.contributor_attributes = contributor_attributes

    @property
    def contributor_orcid(self):
        """Gets the contributor_orcid of this ContributorV20.  # noqa: E501


        :return: The contributor_orcid of this ContributorV20.  # noqa: E501
        :rtype: ContributorOrcidV20
        """
        return self._contributor_orcid

    @contributor_orcid.setter
    def contributor_orcid(self, contributor_orcid):
        """Sets the contributor_orcid of this ContributorV20.


        :param contributor_orcid: The contributor_orcid of this ContributorV20.  # noqa: E501
        :type: ContributorOrcidV20
        """

        self._contributor_orcid = contributor_orcid

    @property
    def credit_name(self):
        """Gets the credit_name of this ContributorV20.  # noqa: E501


        :return: The credit_name of this ContributorV20.  # noqa: E501
        :rtype: CreditNameV20
        """
        return self._credit_name

    @credit_name.setter
    def credit_name(self, credit_name):
        """Sets the credit_name of this ContributorV20.


        :param credit_name: The credit_name of this ContributorV20.  # noqa: E501
        :type: CreditNameV20
        """

        self._credit_name = credit_name

    @property
    def contributor_email(self):
        """Gets the contributor_email of this ContributorV20.  # noqa: E501


        :return: The contributor_email of this ContributorV20.  # noqa: E501
        :rtype: ContributorEmailV20
        """
        return self._contributor_email

    @contributor_email.setter
    def contributor_email(self, contributor_email):
        """Sets the contributor_email of this ContributorV20.


        :param contributor_email: The contributor_email of this ContributorV20.  # noqa: E501
        :type: ContributorEmailV20
        """

        self._contributor_email = contributor_email

    @property
    def contributor_attributes(self):
        """Gets the contributor_attributes of this ContributorV20.  # noqa: E501


        :return: The contributor_attributes of this ContributorV20.  # noqa: E501
        :rtype: ContributorV20
        """
        return self._contributor_attributes

    @contributor_attributes.setter
    def contributor_attributes(self, contributor_attributes):
        """Sets the contributor_attributes of this ContributorV20.


        :param contributor_attributes: The contributor_attributes of this ContributorV20.  # noqa: E501
        :type: ContributorV20
        """

        self._contributor_attributes = contributor_attributes

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ContributorV20, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ContributorV20):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
